from __future__ import annotations

import copy
import re
import types
from typing import Optional, Union
from urllib.parse import urlparse

import pandas as pd

from seeq import spy
from seeq.sdk import *
from seeq.spy import _common
from seeq.spy import _login
from seeq.spy import _url
from seeq.spy import addons
from seeq.spy._errors import *
from seeq.spy._session import Session
from seeq.spy._status import Status
from seeq.spy.addons import _permissions


def install(tool: Union[pd.DataFrame, pd.Series, dict, list], *, include_workbook_parameters: bool = True,
            update_tool: bool = False, update_permissions: bool = False, errors: str = 'raise',
            quiet: bool = False, status: Optional[Status] = None, session: Optional[Session] = None) -> pd.DataFrame:
    """
    Installs or updates Add-on tool(s) in Seeq Workbench.
    Installing Add-on tools requires administrator access.

    Parameters
    ----------
    tool: {dict, list, pd.DataFrame, pd.Series}

        =================== ===================================================
        Property            Description
        =================== ===================================================
        Name                Required. The name of the Add-on tool (shown in
                            green text in the Add-on Tool card).

        Description         Required. The description of the Add-on tool (shown
                            in black text underneath the Name in the tool card).
                            Long descriptions will wrap to the next line and
                            increase the height of the Add-on Tool button.

        Target URL          Required.The URL that the Add-on will open. This
                            can be any website but most commonly a Seeq
                            Data Lab Notebook in Appmode. In order to get the
                            Target URL for a Seeq Data Lab Notebook, open the
                            desired Notebook and copy the URL. In normal mode,
                            the URL contains "/notebooks/" right after the
                            project ID. If you want the notebook to be opened
                            in Appmode, replace "/notebooks/" for "/apps/" in
                            the URL.
                            For example:
                            https://my.seeq.com/data-lab/<id>/apps/TEST.ipynb

        Icon                Optional. Name of the fontawesome icon class to be
                            displayed on the tool card. Defaults to
                            "fa fa-minus". Potential Font Awesome
                            icons can be found at the website
                            https://fontawesome.com/v4.7.0/icons/ and take the
                            form of "fa fa-<icon>" Be sure to include the
                            standalone "fa" in addition to the icon name.

        Link Type           Optional. Sets the display characteristics.
                            Defaults to "window".
                            Options are one of the following:

                            - "window" - display in a new window. After the
                              windows is opened, the user is responsible
                              for subsequent managing of window placement.
                            - "tab" - display in a new tab of the current
                              browser window.
                            - "none" - make a GET request to the URL but do
                              not open a window or tab. Although supported,
                              its use is expected to be rare. It should
                              generally be avoided because it gives the tool
                              card a non-standard button-like behavior.

        Window Details      Optional if Link Type is "window". Sets display
                            characteristics used when Link Type is set to
                            "window". Options are available at
                            https://developer.mozilla.org/en-US/docs/Web/API/Window/open#Window_features.
                            For example:
                            "toolbar=0,height=600,width=600"

        Sort Key            Optional. A string, typically a single character
                            letter. Determines the order in which the Add-on
                            Tools are displayed in the tool panel. The Add-on
                            Tools panel sorts the list of add-ons by this key
                            to determine ordering.

        Reuse Window        Optional. Defaults to False. If True, sets focus
                            to existing window if already opened. Otherwise,
                            open a new window. If False, a new window is opened
                            each time the tool card is clicked. This parameter
                            is only valid when Link Type is set to "window".

        Groups              Optional. List of the Seeq groups that have
                            permission to access the Add-on tool. If the
                            Add-on Tool Target URL is a Data Lab Notebook, then
                            any permissions set on the Add-on Tool will also be
                            added to the target Data Lab Project as users are
                            required to have at least write permissions on the
                            Data Lab Project to run the Notebook. Passing an
                            empty list will remove permissions to the tool for
                            all groups (without modifying permissions to the
                            Target URL). If the tool exists, set
                            update_permissions to True to overwrite the
                            previous Groups permissions. If the tool exists
                            and update_permissions is False, then permissions
                            will not update.

        Users               Optional. List of the Seeq users by username that
                            have permission to access the Add-on tool. To figure
                            out a username, go to the Users tab of the
                            Administration page, click "edit" for the user you
                            want to figure out their username, and get the
                            username from the dialog window that opens. If the
                            Add-on Tool Target URL is a Data Lab Notebook, then
                            any permissions set on the Add-on Tool will also
                            be added to the target Project as users
                            are required to have at least write permissions on
                            the Data Lab Project to run the Notebook. Passing an
                            empty list will remove permissions to the tool for
                            all users (without modifying permissions to the
                            Target URL). If the tool exists, set
                            update_permissions to True to overwrite the
                            previous Users permissions. If the tool exists
                            and update_permissions is False, then permissions
                            will not update.

        ID                  Optional. The ID of an existing Add-on tool. Only
                            needed if you want to update an already installed
                            Add-on tool or its permissions. update_tool or
                            update_permissions must be set to True if you
                            want to update an existing tool.
        =================== ===================================================

    include_workbook_parameters: bool, default True
        If True, the workbookId, worksheetId, workstepId, and/or
        seeqVersion are passed from the launched workbook through the URL as
        query parameters after the “.ipynb?”.
        For example:
        https://my.seeq.com/data-lab/<id>/apps/TEST.ipynb?workbookId={workbookId}&worksheetId={
        worksheetId}&workstepId={workstepId}&seeqVersion={seeqVersion}

    update_tool: bool, default False
        False will prevent an existing Add-on tool to be updated.
        True will update an existing Add-on tool. You must supply the ID of
        the tool(s) that you want to update.

    update_permissions: bool, default False
        False will prevent changing the current permissions on an existing
        Add-on tool and its Target URL.
        True will modify the current permissions on an existing Add-on tool as
        well as its Target URL. You must supply the ID of the tool(s) that you
        want to update its (their) permissions.

    errors : {'raise', 'catalog'}, default 'raise'
        If 'raise', any errors encountered will cause an exception. If
        'catalog', errors will be added to a 'Result' column in the status.df
        DataFrame.

    quiet : bool, default False
        If True, suppresses progress output. Note that when status is
        provided, the quiet setting of the Status object that is passed
        in takes precedent.

    status : spy.Status, optional
        If specified, the supplied Status object will be updated as the command
        progresses. It gets filled in with the same information you would see
        in Jupyter in the blue/green/red table below your code while the
        command is executed. The table itself is accessible as a DataFrame via
        the status.df property.

    session : spy.Session, optional
        If supplied, the Session object (and its Options) will be used to
        store the login session state. This is useful to log in to different
        Seeq servers at the same time or with different credentials.

    Notes
    -----
        Setting the Target URL to a non-appmode Data Lab project should be
        avoided. Non-appmode Data Lab projects would get their own container but
        technical limitations dictate that each container accesses the same
        files on a shared volume. This results in unwanted interactions, save
        notifications, and a last-save-wins behavior.

    Returns
    -------
    pandas.DataFrame
        A DataFrame with the metadata of the Add-on tools pushed, along with
        any errors and statistics about the operation.

        Additionally, the following properties are stored on the "spy"
        attribute of the output DataFrame:

        =================== ===================================================
        Property            Description
        =================== ===================================================
        func                A str value of 'spy.addons.install'
        kwargs              A dict with the values of the input parameters
                            passed to spy.addons.install to get the output
                            DataFrame
        status              A spy.Status object with the status of the
                            spy.addons.install call
        =================== ===================================================

    Examples
    --------
    Install a new Add-on tool
    >>> my_new_tool = {
    >>>     "Name": 'My New Tool',
    >>>     "Description": "this is an awesome tool",
    >>>     "Target URL": "https://www.my.seeq.com/data-lab/8FB008FF-ECF0-4837-B3D4-7FCBB49CC108/apps/my_tool.ipynb",
    >>>     "Groups": ['My Seeq Group']}

    >>> df_installed = spy.addons.install(my_new_tool)

    If the tool already existed and you want to update it
    >>> df_installed = spy.addons.install(my_new_tool, update_tool=True)

    To installed multiple tools at once:
    >>> new_tools = [
    >>>     {
    >>>         "Name": 'My New Tool',
    >>>         "Description": "this is an awesome tool",
    >>>         "Target URL": "https://my.seeq.com/data-lab/8FB008FF-ECF0-4837-B3D4-7FCBB49CC108/apps/my_tool.ipynb",
    >>>         "Icon": "fa fa-bell",
    >>>         "Groups": ['My Seeq Group']},
    >>>     {
    >>>         "Name": 'Your New Tool',
    >>>         "Description": "this is an awesome tool",
    >>>         "Target URL": "https://my.seeq.com/data-lab/F4DS345D-23DS-2344-SVSE-345234TSDF52/apps/tool2.ipynb",
    >>>         "Icon": "fa fa-bars",
    >>>         "Groups": ['Everyone']}]

    >>> df_installed = spy.addons.install(new_tools)

    Or set the properties in a pd.DataFrame (useful when modifying existing tools by manipulating the pd.DataFrame
    from spy.addons.search):

    >>> searched_tools = spy.addons.search({"Link Type": "window"})
    >>> searched_tools["Link Type"] = ['tab'] * len(searched_tools)
    >>> df_installed = spy.addons.install(searched_tools,update_tool=True)

    To change group permissions:
    >>> searched_tools = spy.addons.search({"Name": "My New Tool"})
    >>> searched_tools["Groups"] = [['Everyone']] * len(searched_tools)
    >>> df_installed = spy.addons.install(searched_tools,update_permissions=True)

    To change user permissions
    >>> searched_tools = spy.addons.search({"Name": "My New Tool"})
    >>> searched_tools["Users"] = [['user@my.seeq.com']] * len(searched_tools)
    >>> df_installed = spy.addons.install(searched_tools,update_permissions=True)

    """

    input_args = _common.validate_argument_types([
        (tool, 'tool', (str, dict, list, pd.DataFrame, pd.Series)),
        (include_workbook_parameters, 'include_workbook_parameters', bool),
        (update_tool, "update_tool", bool),
        (update_permissions, "update_permissions", bool),
        (errors, 'errors', str),
        (quiet, 'quiet', bool),
        (status, 'status', Status),
        (session, 'session', Session)
    ])

    status = Status.validate(status, quiet)
    session = Session.validate(session)
    _login.validate_login(session, status)

    try:
        return _install(session, tool, include_workbook_parameters=include_workbook_parameters,
                        update_tool=update_tool,
                        update_permissions=update_permissions, errors=errors, quiet=quiet, status=status,
                        input_args=input_args)

    except KeyboardInterrupt:
        status.update('Add-on installation canceled', Status.CANCELED)


def _install(session: Session, tool, *, include_workbook_parameters=True, update_tool=False, update_permissions=False,
             errors='raise', quiet=False, status=None, input_args=None):
    status = Status.validate(status, quiet)
    tool_copy = copy.deepcopy(tool)  # To avoid modifying the original object passed by the user
    system_api = SystemApi(session.client)

    if isinstance(tool_copy, pd.DataFrame):
        tools = tool_copy.to_dict(orient='records')
    elif isinstance(tool_copy, pd.Series):
        tools = [tool_copy.to_dict()]
    elif isinstance(tool_copy, list):
        tools = tool_copy
    else:
        tools = [tool_copy]

    for _tool in tools:
        tool_types = []
        for k, v in _tool.items():
            if k == "Reuse Window":
                tool_types.append((v, k, bool))
            elif k in ['Groups', 'Users']:
                tool_types.append((v, k, list))
                if isinstance(v, list):
                    tool_types.extend([(x, str(x), str) for x in v])
            elif k in ['Name', 'Description', 'Target URL', 'Icon', 'Link Type', 'Window Details', 'Sort Key', 'ID']:
                tool_types.append((v, k, str))
                if k == 'ID':
                    if not _common.is_guid(v):
                        raise SPyValueError(f'The ID "{v}" is malformed')
            elif k in ['Effective Permissions', 'Type', 'Archived', 'Status Message']:
                pass  # Ignore these keys since they might be coming from spy.addons.search
            else:
                raise SPyValueError(f"Property {k} is not a valid option for the `tool` parameter")

        _common.validate_argument_types(tool_types)

    defaults = {'Icon': 'fa fa-minus',
                'Link Type': 'window',
                'Sort Key': 'a',
                'Reuse Window': False,
                'Window Details': ''}

    required_props = {"Name", "Description", "Target URL"}
    tools_old = system_api.get_add_on_tools().add_on_tools

    status.df = pd.DataFrame()
    status.df['ID'] = ''
    status.df['Name'] = ''
    status.df['Target URL'] = ''
    status.df['Time'] = 0
    status.df['Result'] = 'Queued'
    status.update('Installing', Status.RUNNING)

    for status_idx, _tool in enumerate(tools):
        timer = _common.timer_start()
        tool_id = None

        if update_tool or update_permissions:
            status.update(f"Updating Add-on tool", Status.RUNNING)
            if 'ID' not in _tool.keys() and 'Name' not in _tool.keys():
                message = f"Either the Name or the ID must be supplied when update_tool is True"
                _common.raise_or_catalog(errors, status=status, exception_type=SPyException, index=status_idx,
                                         message=message)
                continue

            if 'ID' in _tool.keys():
                tool_id = _tool['ID']

            if 'ID' not in _tool.keys():
                tools_with_same_name = [x for x in tools_old if x.name == _tool['Name']]
                if len(tools_with_same_name) > 1:
                    message = f'There exists {len(tools_with_same_name)} tools with the name "{_tool["Name"]}". ' \
                              f'You can supply the ID to avoid ambiguity and modify only the intended tool'
                    _common.raise_or_catalog(errors, status=status, exception_type=SPyException, index=status_idx,
                                             message=message)
                elif len(tools_with_same_name) == 1:
                    tool_id = tools_with_same_name[0].id

            if tool_id is not None:
                tool_record = addons.search({"ID": tool_id}, quiet=True, errors='catalog', session=session) \
                    .to_dict('records')
                if not tool_record:
                    message = f'A tool with ID "{tool_id}" does not exist'
                    _common.raise_or_catalog(errors, status=status, exception_type=SPyException, index=status_idx,
                                             message=message)
                tool_old = tool_record[0]
                # Update the properties set by the user
                old_permissions = _permissions.get_addon_permissions(tool_id, session=session)
                tool_old['Groups'] = old_permissions['Groups']
                tool_old['Users'] = old_permissions['Users']
                for k in tool_old.keys():
                    if k not in _tool:
                        _tool[k] = tool_old[k]

        if tool_id is None:
            status.update('Installing Add-on', Status.RUNNING)
            if not required_props.issubset(_tool.keys()):
                addendum = ''
                if _tool.get('ID') is not None:
                    addendum = "Consider setting `update_tool=True` if your intention is to update only certain " \
                               "properties of an existing the tool."
                message = f"The properties {required_props} are required. Got properties {set(_tool.keys())} for tool" \
                          f" {status_idx}. {addendum}"
                _common.raise_or_catalog(errors, status=status, exception_type=SPyException, index=status_idx,
                                         message=message)
                continue
            # Use defaults for properties not set by the user
            if 'Icon' not in _tool.keys():
                _tool['Icon'] = defaults['Icon']

            if 'Link Type' not in _tool.keys():
                _tool['Link Type'] = defaults['Link Type']

            if 'Window Details' not in _tool.keys():
                _tool['Window Details'] = defaults['Window Details']

            if "Sort Key" not in _tool.keys():
                _tool['Sort Key'] = defaults['Sort Key']

            if "Reuse Window" not in _tool.keys():
                _tool['Reuse Window'] = defaults['Reuse Window']

            if 'ID' not in _tool.keys():
                _tool['ID'] = None

        for prop in ['Groups', 'Users']:
            if prop not in _tool.keys():
                _tool[prop] = []

        parsed_url = urlparse(_tool['Target URL'])
        target_url = parsed_url.geturl()

        datalab_project_id = _url.get_data_lab_project_id_from_url(target_url)
        if datalab_project_id is None:
            status.warn(
                f'It looks like the target URL for tool "{_tool["Name"]}" is not a Seeq Data Lab project. This is '
                f'typically not recommended. Make sure this is indeed your intent')

        query_params = parsed_url.query
        if include_workbook_parameters:
            workbook_params = '&workbookId={workbookId}&worksheetId={worksheetId}&workstepId={workstepId}' \
                              '&seeqVersion={seeqVersion}'

            if not re.search(workbook_params[1:], query_params):
                query_params += workbook_params
            query_params = query_params[1:] if query_params[0] == '&' else query_params

        else:
            # In case the workbook query parameters exists but user passes include_workbook_parameters=False
            for params in ['workbookId={workbookId}', 'worksheetId={worksheetId}', 'workstepId={workstepId}',
                           'seeqVersion={seeqVersion}']:
                query_params = query_params.replace(params, '')
                query_params = query_params.replace('&&', '&')
            if query_params:
                query_params = query_params[1:] if query_params[0] == '&' else query_params
            if query_params:
                query_params = query_params[:-1] if query_params[-1] == '&' else query_params

        new_tool_config = dict(
            name=_tool['Name'],
            description=_tool['Description'],
            iconClass=_tool['Icon'],
            targetUrl=f'{parsed_url.scheme}://{parsed_url.netloc}{parsed_url.path}?{query_params}',
            linkType=_tool['Link Type'],
            windowDetails=_tool['Window Details'],
            sortKey=_tool['Sort Key'],
            # need to cast to bool if coming from pandas as np.bool_
            reuseWindow=True if _tool['Reuse Window'] else False,
            permissions=dict(groups=_tool['Groups'],
                             users=_tool['Users'])
        )

        status.df.at[status_idx, 'Name'] = new_tool_config['name']
        status.df.at[status_idx, 'Target URL'] = new_tool_config['targetUrl']

        if not update_tool and update_permissions and _tool.get('ID'):
            try:
                # We came this far for this check since we still need the target URL and permissions default checks
                status.df.at[status_idx, 'ID'] = _tool['ID']
                status.df.at[status_idx, 'Time'] = _common.timer_elapsed(timer)
                _permissions.set_permissions(session, new_tool_config, _tool['ID'])
                status.df.at[status_idx, 'Result'] = 'Updated permissions'
            except (ApiException, AssertionError, SPyValueError, SPyRuntimeError) as e:
                status.df.at[status_idx, 'Time'] = _common.timer_elapsed(timer)
                _common.raise_or_catalog(errors, status=status, index=status_idx, exception_type=SPyException,
                                         message=e)
        else:
            new_id = None
            try:
                new_id = _install_or_update_tool(session, new_tool_config, update_tool=update_tool, tool_id=_tool['ID'])
                if update_permissions or _tool['ID'] is None:
                    _permissions.set_permissions(session, new_tool_config, new_id)
                status.df.at[status_idx, 'ID'] = new_id
                status.df.at[status_idx, 'Time'] = _common.timer_elapsed(timer)
                status.df.at[status_idx, 'Result'] = f"{'Updated' if _tool.get('ID') else 'Installed'}"
            except (ApiException, AssertionError, SPyValueError, SPyRuntimeError) as e:
                if new_id is not None and _tool['ID'] is None:
                    # We should get here only if there was a NEW tool that failed to be created
                    system_api.delete_add_on_tool(id=new_id)
                status.df.at[status_idx, 'Time'] = _common.timer_elapsed(timer)
                _common.raise_or_catalog(errors, status=status, index=status_idx, exception_type=SPyException,
                                         message=e)

    status.update(f'Add-on tool{"s" if len(tools) > 1 else ""} installation successful', Status.SUCCESS)

    install_result_df = status.df.copy()
    push_df_properties = types.SimpleNamespace(
        func='spy.addons.install',
        kwargs=input_args,
        status=status)

    _common.put_properties_on_df(install_result_df, push_df_properties)

    return install_result_df


def _install_or_update_tool(session: Session, new_tool_config, update_tool, tool_id):
    system_api = SystemApi(session.client)
    tools = system_api.get_add_on_tools().add_on_tools

    # If the tool is in the list
    tool_names = [tool.name for tool in tools]
    if new_tool_config["name"] in tool_names:
        if not update_tool or tool_id is None:
            message = f'Add-on tool with name "{new_tool_config["name"]}" already exists. You can update the ' \
                      f'existing tool with `update_tool=True`'
            raise SPyRuntimeError(message)

    tool = new_tool_config.copy()
    tool.pop("permissions")
    if tool_id is not None:
        # Update tool
        system_api.update_add_on_tool(id=tool_id, body=tool)
    else:
        # Create add-on tool
        tool_id = system_api.create_add_on_tool(body=tool).id

    return tool_id



from __future__ import annotations

import types
from typing import Optional, Union, Hashable

import pandas as pd

from seeq import spy
from seeq.sdk import *
from seeq.spy import _common
from seeq.spy import _login
from seeq.spy._errors import *
from seeq.spy._redaction import request_safely
from seeq.spy._session import Session
from seeq.spy._status import Status


def uninstall(items: Union[pd.DataFrame, pd.Series], *, errors: str = 'raise', quiet: bool = False,
              status: Status = None, session: Optional[Session] = None) -> pd.DataFrame:
    """
    Uninstalls Add-on Tool from the Seeq Workbench. It does not remove the
    target_url contents. Uninstalling Add-on tools requires administrator access.

    Parameters
    ----------
    items : {pd.DataFrame, pd.Series}
        A DataFrame or Series containing ID column that can be used to
        identify the Add-on tool to uninstall. This is usually created via a call
        to spy.addons.search().

    errors : {'raise', 'catalog'}, default 'raise'
        If 'raise', any errors encountered will cause an exception. If
        'catalog', errors will be added to a 'Result' column in the status.df
        DataFrame.

    quiet : bool
        If True, suppresses progress output. Note that when status is
        provided, the quiet setting of the Status object that is passed
        in takes precedent.

    status : spy.Status, optional
        If specified, the supplied Status object will be updated as the command
        progresses. It gets filled in with the same information you would see
        in Jupyter in the blue/green/red table below your code while the
        command is executed. The table itself is accessible as a DataFrame via
        the status.df property.

    session : spy.Session, optional
        If supplied, the Session object (and its Options) will be used to
        store the login session state. This is useful to log in to different
        Seeq servers at the same time or with different credentials.

    Returns
    -------
    pandas.DataFrame
        A DataFrame with the metadata of the Add-on tools uninstalled, along with
        any errors and statistics about the operation.

        Additionally, the following properties are stored on the "spy"
        attribute of the output DataFrame:

        =================== ===================================================
        Property            Description
        =================== ===================================================
        func                A str value of 'spy.addons.uninstall'
        kwargs              A dict with the values of the input parameters
                            passed to spy.addons.uninstall to get the output
                            DataFrame
        status              A spy.Status object with the status of the
                            spy.addons.uninstall call
        =================== ===================================================

    Examples
    --------
    Search for a tool with a specific name and uninstall it

    >>> search_results = spy.addons.search({'Name': 'Obsolete Tool'})
    >>> uninstalled_tool = spy.addons.uninstall(search_results)

    """

    input_args = _common.validate_argument_types([
        (items, 'items', (pd.DataFrame, pd.Series)),
        (errors, 'errors', str),
        (quiet, 'quiet', bool),
        (status, 'status', Status),
        (session, 'session', Session)
    ])

    status = Status.validate(status, quiet)
    session = Session.validate(session)
    _login.validate_login(session, status)
    system_api = SystemApi(session.client)

    _common.validate_errors_arg(errors)

    if 'ID' not in items:
        raise SPyValueError('items DataFrame must include "ID" column which is typically obtained from '
                            '"spy.addons.search"')
    if isinstance(items, pd.Series):
        items = pd.DataFrame().append(items)
    if not isinstance(items, pd.DataFrame):
        raise SPyTypeError('item must be a pandas.Series or pandas.DataFrame')

    status_columns = [c for c in ['ID', 'Name', 'Target URL'] if c in items]

    status.df = items[status_columns].copy()
    status.df['Count'] = 0
    status.df['Time'] = 0
    status.df['Result'] = 'Uninstalling'
    timer = _common.timer_start()

    for row_index, row in items.iterrows():
        @request_safely(action_description=f"uninstalling addon with id {row['ID']}", errors=errors, status=status)
        def uninstall_addon(index: Hashable, df_row: pd.Series):
            system_api.delete_add_on_tool(id=df_row['ID'])
            status.df.at[index, 'Time'] = _common.timer_elapsed(timer)
            status.df.at[index, 'Count'] = 1
            status.df.at[index, 'Result'] = 'Uninstalled'

        uninstall_addon(row_index, row)

    status.update(f'Add-on tool{"s" if len(items) > 1 else ""} uninstall successful', Status.SUCCESS)

    uninstall_result_df = status.df.copy()
    push_df_properties = types.SimpleNamespace(
        func='spy.addons.uninstall',
        kwargs=input_args,
        status=status)

    _common.put_properties_on_df(uninstall_result_df, push_df_properties)

    return uninstall_result_df



from __future__ import annotations

import math
import string
import textwrap
from enum import Enum
from typing import Callable, Dict, Mapping, Optional, Union, List

import pandas as pd
from deprecated import deprecated

from seeq.spy import _common
from seeq.spy import _pull
from seeq.spy import _push
from seeq.spy._errors import *
from seeq.spy._session import Session
from seeq.spy.assets._context import BuildContext, BuildMode, BuildPhase, SPyInstanceAlreadyExists
from seeq.spy.workbooks import Analysis, AnalysisWorksheet, Topic, TopicDocument, DateRange


class WorkbookNotBuilt(SPyException):
    def __init__(self, workbook_type, workbook, worksheet=None):
        self.workbook_type = workbook_type
        self.workbook = workbook
        self.worksheet = worksheet


METHOD_TYPE_ATTR = 'spy_model'
FRIENDLY_NAME_ATTR = 'spy_friendly_name'


class MethodType(Enum):
    ATTRIBUTE = 'Attributes'
    COMPONENT = 'Components'
    DISPLAY = 'Displays'
    DATE_RANGE = 'Date Ranges'
    DOCUMENT = 'Documents'
    PLOT = 'Plots'
    REQUIREMENT = 'Requirements'


class PlotRenderInfo:

    def __init__(self, image_format, render_function):
        self.image_format = image_format
        self.render_function = render_function


def get_method_friendly_name(func: Callable):
    return func.__name__.replace('_', ' ')


class _AssetBase:
    definition: Union[dict, Mapping]
    context: BuildContext
    exceptions: Dict[str, str]

    DEFAULT_WORKBOOK_NEEDED = '__DEFAULT_WORKBOOK_NEEDED__'
    UNKNOWN = '__unknown__'

    def __init__(self, context: BuildContext, definition: Union[dict, pd.DataFrame, pd.Series] = None, *, parent=None):
        """
        Instantiates an Asset or Mixin.

        :param context: A BuildContext object used to store information about the build operation and all other
                        assets in the build.
        :type context: BuildContext
        :param definition: A dictionary of property-value pairs that make up the definition of the Asset or Mixin.
                           Typically you will want to supply 'Name' at minimum.
        :type definition: dict, pd.DataFrame, pd.Series
        :param parent: An instance of either an Asset or Mixin that represents the parent of this instance. Typically
                       this is supplied when @Asset.Component is used to define child assets.
        :type parent: Asset
        """
        self.definition = dict()
        self.exceptions = dict()

        if isinstance(definition, _AssetBase):
            self.definition = definition.definition
        elif isinstance(definition, pd.DataFrame):
            if len(definition) != 1:
                raise SPyValueError('DataFrame must be exactly one row')
            self.definition = definition.iloc[0].to_dict()
        elif isinstance(definition, pd.Series):
            self.definition = definition.to_dict()
        elif definition is not None:
            self.definition = definition

        if _common.present(self.definition, 'Path'):
            self.definition['Path'] = _common.sanitize_path_string(self.definition['Path'])

        self.definition['Type'] = 'Asset'
        if 'Name' in self.definition:
            # For an Asset, its name and the Asset column are made identical for clarity
            self.definition['Asset'] = self.definition['Name']

        self.context = context

        self._parent = self.UNKNOWN  # This is a cache for the parent once it is found

        if parent is not None and parent != self.UNKNOWN:
            self._parent = parent

            # Passing in a parent will relieve the user from having to construct the right path
            if _common.present(parent.definition, 'Path'):
                self.definition['Path'] = parent.definition['Path'] + ' >> ' + parent.definition['Name']
            else:
                self.definition['Path'] = parent.definition['Name']

        self.definition['Asset Object'] = self

        # 'Template' is set on the asset with the hope that, in the future, we will be able to search for items in
        # the asset tree that are derived from a particular template.
        self.definition['Template'] = self.template_friendly_name

        self.initialize()

        if not self.is_mixin:
            # Register ourselves in the build context
            context.add_object(self)

    def initialize(self):
        # This is for users to override so they don't have to know about the "crazy" __init__ function and its arguments
        pass

    @property
    def is_mixin(self):
        return issubclass(self.__class__, Mixin)

    def __contains__(self, key):
        return _common.present(self.definition, key)

    def __getitem__(self, key):
        return _common.get(self.definition, key)

    def __setitem__(self, key, val):
        self.definition[key] = val

    def __delitem__(self, key):
        del self.definition[key]

    def __repr__(self):
        return self.fqn

    @property
    @deprecated(reason="Use self.definition instead")
    def asset_definition(self):
        return self.definition

    @property
    @deprecated(reason="Use self.parent.definition instead")
    def parent_definition(self):
        return self.parent.definition if self.parent is not None else None

    @property
    def fqn(self):
        """
        The Fully-Qualified Name of this object, which includes both the Path and the Name using the usual >>
        separators.
        """
        return '%s >> %s' % (_common.get(self, 'Path'), _common.get(self, 'Name'))

    @property
    def parent(self):
        """
        The parent Asset object of this Asset or Mixin, if it exists.
        """
        if self._parent != self.UNKNOWN:
            return self._parent

        expected_parent_path_list = _common.path_string_to_list(self.fqn)
        if len(expected_parent_path_list) == 0:
            self._parent = None
        else:
            expected_parent_path_list = expected_parent_path_list[0:-1]
            for asset in self.all_assets():
                if _common.path_string_to_list(asset.fqn) == expected_parent_path_list:
                    self._parent = asset
                    break

        return self._parent

    def all_assets(self):
        """
        All asset instances in the entire tree.

        Note that, during the INSTANTIATING build phase (the first), this function returns
        a blank list because not all asset objects have been instantiated yet.

        :return: A list of all asset instances in the entire tree.
        """
        if self.context.phase == BuildPhase.INSTANTIATING:
            # Since the context is currently being filled in with objects, it's better to return an empty list here
            # rather than confusingly returning a half-filled-in list.
            return list()

        return self.context.objects.values()

    def is_child_of(self, asset):
        """
        Tests if this asset instance (self) is a direct child of the specified asset.
        :param asset: The asset that might be self's parent
        :return: True if self is a direct child of asset
        """
        return self.parent == asset

    def is_parent_of(self, asset):
        """
        Tests if this asset instance (self) is a direct parent of the specified asset.
        :param asset: The asset that might be self's child
        :return: True if self is a direct parent of asset
        """
        return self == asset.parent

    def is_ancestor_of(self, asset):
        """
        Tests if this asset instance (self) is a parent or grandparent etc of the specified asset (looking all the way
        up the tree).
        :param asset: The asset that might be above self in the hierarchy
        :return: True if self is above asset
        """
        return asset.fqn.startswith(self.fqn) and asset.fqn != self.fqn

    def is_descendant_of(self, asset):
        """
        Tests if this asset instance (self) is a child or grandchild etc of the specified asset (looking all the way
        down the tree).
        :param asset: The asset that might be below self in the hierarchy
        :return: True if self is below asset
        """
        return self.fqn.startswith(asset.fqn) and asset.fqn != self.fqn

    def get_model_methods(self, method_type: MethodType = None):
        # Filter out property members so that the Deprecated library doesn't produce warnings as we iterate over them
        method_names = [m for m in dir(self) if m not in ['asset_definition', 'parent_definition', 'parent', 'fqn',
                                                          'is_mixin', 'template_name', 'template_friendly_name']]

        # Assemble a list of all functions on this object instance that are callable so that we can iterate over them
        # and find @Asset.Attribute() and @Asset.Component() functions
        object_methods = [getattr(self, method_name) for method_name in method_names
                          if callable(getattr(self, method_name))]

        # The "spy_model" attribute is added to any @Asset.Attribute() and @Asset.Component() decorated
        # functions so that they are processed during build
        remaining_methods = [func for func in object_methods if hasattr(func, METHOD_TYPE_ATTR)]

        if method_type:
            remaining_methods = filter(lambda m: getattr(m, METHOD_TYPE_ATTR) == method_type, remaining_methods)

        return remaining_methods

    def build(self, metadata: pd.DataFrame):
        definitions = list()
        context = self.context

        def _include_method(_func):
            # The build is made in two passes, first to instantiate the objects and then to build all the attributes,
            # requirements, documents, date ranges etc. Components must be touched in both phases because they
            # can instantiate objects.

            _func_type = getattr(_func, METHOD_TYPE_ATTR)
            if self.context.phase == BuildPhase.INSTANTIATING:
                return _func_type in [MethodType.COMPONENT, MethodType.REQUIREMENT]
            else:
                return _func_type not in [MethodType.REQUIREMENT]

        remaining_methods = [m for m in self.get_model_methods() if _include_method(m)]

        while len(remaining_methods) > 0:
            at_least_one_built = False
            at_least_one_dependency_not_built = False
            workbooks_not_built = list()

            for func in remaining_methods.copy():
                func_type = getattr(func, 'spy_model')

                # noinspection PyBroadException
                try:
                    func_results = func(metadata)

                except WorkbookNotBuilt as e:
                    workbooks_not_built.append(e)
                    continue

                except SPyDependencyNotBuilt as e:
                    at_least_one_dependency_not_built = True
                    self.exceptions[getattr(func, FRIENDLY_NAME_ATTR)] = \
                        f'"{SPyAssetMemberError.get_friendly_path_for_error(self, func)}": {str(e)}'
                    continue

                except SPyAssetMemberError as e:
                    self.exceptions[getattr(func, FRIENDLY_NAME_ATTR)] = str(e)
                    continue

                except BaseException:
                    self.exceptions[getattr(func, FRIENDLY_NAME_ATTR)] = _common.format_exception()
                    continue

                if getattr(func, FRIENDLY_NAME_ATTR) in self.exceptions:
                    del self.exceptions[getattr(func, FRIENDLY_NAME_ATTR)]

                at_least_one_built = True
                remaining_methods.remove(func)

                if func_results is None or func_type not in [MethodType.ATTRIBUTE,
                                                             MethodType.COMPONENT,
                                                             MethodType.REQUIREMENT]:
                    continue

                if isinstance(func_results, list):
                    # This is the @Asset.Component case
                    definitions.extend(func_results)
                elif isinstance(func_results, dict):
                    # This is the @Asset.Attribute case
                    definitions.append(func_results)

            if not at_least_one_built and len(workbooks_not_built) > 0:
                for not_built in workbooks_not_built:
                    if (not_built.workbook_type, not_built.workbook) not in self.context.workbooks:
                        if not_built.workbook_type == 'analysis':
                            workbook = Analysis({'Name': not_built.workbook})
                        else:
                            workbook = Topic({'Name': not_built.workbook})
                        self.context.workbooks[(not_built.workbook_type, not_built.workbook)] = workbook

                at_least_one_built = True

            if not at_least_one_built:
                if at_least_one_dependency_not_built:
                    raise SPyDependencyNotBuilt(self, None, 'at least one asset member dependency not built')

                break

        if context.phase == BuildPhase.BUILDING and not self.is_mixin:
            definitions.append(self.definition)
            self.context.add_results(self.definition)

        self.definition['Build Result'] = BuildPhase.SUCCESS

        return definitions

    def build_component(self, template, metadata, component_name) -> List[dict]:
        """
        Builds a single component by instantiating the supplied template
        and building it with the supplied metadata.

        Parameters
        ----------
        template : {Asset}
            An asset class that is used to construct the attributes and sub-
            components.

        metadata : pd.DataFrame
            The metadata DataFrame containing all rows relevant to the
            sub-component.

        component_name : str
            The name of the sub-component. This will be used as the Name
            property of the instantiated asset (unless template is a Mixin.)

        Returns
        -------
        list(dict)
            A list of definitions (comprising one item). Note that this list
            will be empty in the first build phase (where self.context.phase
            equals BuildPhase.INSTANTIATING), as only the asset objects will
            have been instantiated.
        """

        try:
            instance = template(self.context, {
                'Name': component_name,
            }, parent=self)  # type: _AssetBase
        except SPyInstanceAlreadyExists as e:
            instance = e.instance

        exception_key = f'[Component] {component_name}'

        def _propagate_exceptions():
            if len(instance.exceptions) > 0:
                self.exceptions[exception_key] = '\n'.join(instance.exceptions.values())

        try:
            component_definition = instance.build(metadata)
            if exception_key in self.exceptions:
                del self.exceptions[exception_key]
        except BaseException:
            _propagate_exceptions()
            raise

        _propagate_exceptions()

        return component_definition

    def build_components(self, template, metadata: pd.DataFrame, column_name: str) -> List[dict]:
        """
        Builds a set of components by identifying the unique values in the
        column specified by column_name and then instantiating the supplied
        template for each one and building it with the subset of metadata
        for that column value.

        Useful when constructing a rich model whereby a root asset is composed
        of unique components, possibly with further sub-components. For
        example, you may have a Plant asset that contains eight Refigerator
        units that each have two associated Compressor units.

        Parameters
        ----------
        template : {Asset}
            An asset class that is used to construct the attributes and sub-
            components.

        metadata : pd.DataFrame
            The metadata DataFrame containing all rows relevant to all
            (sub-)components of this asset. The DataFrame must contain the
            column specified by column_name.

        column_name : str
            The name of the column that will be used to discover the unique
            (sub-)components of this asset. For example, if column_name=
            'Compressor', then there might be values of 'Compressor A12' and
            'Compressor B74' in the 'Compressor' column of the metadata
            DataFrame.

        Returns
        -------
        list(dict)
            A list of definitions for each component. Note that this list
            will be empty in the first build phase (where self.context.phase
            equals BuildPhase.INSTANTIATING), as only the asset objects will
            have been instantiated.

        Examples
        --------
        Define a Refrigerator template that has Compressor subcomponents.

        >>> class Refrigerator(Asset):
        >>>     @Asset.Attribute()
        >>>     def Temperature(self, metadata):
        >>>         return metadata[metadata['Name'].str.endswith('Temperature')]
        >>>
        >>>     @Asset.Component()
        >>>     def Compressor(self, metadata):
        >>>         return self.build_components(Compressor, metadata, 'Compressor')
        >>>
        >>> class Compressor(Asset):
        >>>
        >>>     @Asset.Attribute()
        >>>     def Power(self, metadata):
        >>>         return metadata[metadata['Name'].str.endswith('Power')]
        """
        component_names = metadata[column_name].dropna().drop_duplicates().tolist()
        component_definitions = ItemGroup()
        at_least_one_dependency_not_built = False
        for component_name in component_names:
            component_metadata = metadata[metadata[column_name] == component_name]
            try:
                component_definition = self.build_component(template, component_metadata, component_name)
                component_definitions.extend(component_definition)
            except SPyDependencyNotBuilt:
                at_least_one_dependency_not_built = True

        if at_least_one_dependency_not_built:
            raise SPyDependencyNotBuilt(self, None, 'at least one asset member dependency not built')

        return component_definitions

    def pull(self, items, *, start=None, end=None, grid='15min', header='__auto__', group_by=None,
             shape='auto', capsule_properties=None, tz_convert=None, calculation=None, bounding_values=False,
             errors='raise', session: Optional[Session] = None):
        session = Session.validate(session)

        if isinstance(items, list):
            items = pd.DataFrame(items)

        _common.validate_unique_dataframe_index(items, 'items')

        for index, item in items.iterrows():
            if not _common.present(item, 'ID') or _common.get(item, 'Reference', False):
                pushed_item = _push.get_from_push_df(item, self.context.push_df)
                items.at[index, 'ID'] = pushed_item['ID']

        return _pull.pull(pd.DataFrame(items), start=start, end=end, grid=grid, header=header, group_by=group_by,
                          shape=shape, capsule_properties=capsule_properties, tz_convert=tz_convert,
                          calculation=calculation, bounding_values=bounding_values, errors=errors,
                          status=self.context.status, session=session)

    @staticmethod
    def _add_asset_metadata(asset, attribute_definition):
        if 'Path' in asset.definition and 'Path' not in attribute_definition:
            attribute_definition['Path'] = asset.definition['Path']

        if 'Asset' in asset.definition and 'Asset' not in attribute_definition:
            attribute_definition['Asset'] = asset.definition['Asset']

        if 'Template' in asset.definition and 'Template' not in attribute_definition:
            attribute_definition['Template'] = asset.template_friendly_name

    @staticmethod
    def _build_attribute_definition(asset, func, func_results):
        attribute_definition = dict()

        def _preserve_originals():
            for _key in ['Name', 'Path', 'Asset', 'Datasource Class', 'Datasource ID', 'Data ID',
                         'Source Number Format', 'Source Maximum Interpolation', 'Source Value Unit Of Measure']:
                if _common.present(attribute_definition, _key):
                    attribute_definition['Referenced ' + _key] = attribute_definition[_key]
                    del attribute_definition[_key]

        if isinstance(func_results, pd.DataFrame):
            if len(func_results) == 1:
                attribute_definition.update(func_results.iloc[0].to_dict())
                _preserve_originals()
                attribute_definition['Reference'] = True
            elif len(func_results) > 1:
                raise SPyAssetMemberError(asset, func, f'Multiple attributes returned:\n{func_results}')
            else:
                raise SPyAssetMemberError(asset, func, 'No matching metadata row found')

        elif isinstance(func_results, dict):
            attribute_definition.update(func_results)
            if _common.present(func_results, 'ID'):
                # If the user is supplying an identifier, they must intend it to be a reference, otherwise
                # it can't be in the tree.
                attribute_definition['Reference'] = True

        if not _common.present(attribute_definition, 'Name'):
            attribute_definition['Name'] = get_method_friendly_name(func)

        attribute_definition['Asset'] = asset.definition['Name']
        attribute_definition['Asset Object'] = asset

        _AssetBase._add_asset_metadata(asset, attribute_definition)

        if 'Formula Parameters' in attribute_definition:
            for key, val in attribute_definition['Formula Parameters'].items():
                if val is None:
                    raise SPyAssetMemberError(
                        asset, func,
                        f'Formula Parameter "{key}" not found. If necessary, check for None before attempting to use '
                        'optional Requirements or Attribute dependencies and return None from your Attribute function '
                        'to exclude the calculated attribute from your asset (if desired).')

        for key in ['Formula', 'Description']:
            if key in attribute_definition and isinstance(attribute_definition[key], str):
                attribute_definition[key] = textwrap.dedent(attribute_definition[key]).strip()

        return attribute_definition

    @property
    def template_name(self):
        return self.__class__.__name__

    @property
    def template_friendly_name(self):
        return self.template_name.replace('_', ' ')

    @staticmethod
    def _set_wrapper_attrs(wrapper: Callable, method_type: MethodType, func: Callable):
        # Setting this attribute on the function itself makes it discoverable during build()
        setattr(wrapper, METHOD_TYPE_ATTR, method_type)

        setattr(wrapper, FRIENDLY_NAME_ATTR, get_method_friendly_name(func))
        setattr(wrapper, '__doc__', func.__doc__)

    # noinspection PyPep8Naming
    @classmethod
    def Attribute(cls):
        """
        This decorator appears as @Asset.Attribute on a function with a class that derives from Asset.
        """

        def attribute_decorator(func):
            def attribute_wrapper(self, metadata=None):
                # type: (_AssetBase, pd.DataFrame) -> Optional[dict]
                if (self, func.__name__) in self.context.cache:
                    return self.context.cache[(self, func.__name__)]

                if metadata is None:
                    raise SPyDependencyNotBuilt(self, func, 'attribute dependency not built')

                func_results = func(self, metadata)

                attribute_definition = None
                if func_results is not None:
                    if not isinstance(func_results, list):
                        attribute_definition = _AssetBase._build_attribute_definition(self, func, func_results)
                        attribute_definition['Build Result'] = BuildPhase.SUCCESS
                    else:
                        attribute_definition = ItemGroup()
                        for func_result in func_results:
                            individual_attribute_def = _AssetBase._build_attribute_definition(self, func, func_result)
                            individual_attribute_def['Build Result'] = BuildPhase.SUCCESS
                            attribute_definition.append(individual_attribute_def)

                self.context.cache[(self, func.__name__)] = attribute_definition
                self.context.add_results(attribute_definition)
                self.context.at_least_one_thing_built_somewhere = True

                return attribute_definition

            _AssetBase._set_wrapper_attrs(attribute_wrapper, MethodType.ATTRIBUTE, func)

            return attribute_wrapper

        return attribute_decorator

    # noinspection PyPep8Naming
    @classmethod
    def Component(cls):
        """
        This decorator appears as @Asset.Component on a function with a class that derives from Asset.
        """

        def component_decorator(func):
            def component_wrapper(self, metadata=None):
                # type: (_AssetBase, pd.DataFrame) -> ItemGroup
                if (self, func.__name__) in self.context.cache:
                    return self.context.cache[(self, func.__name__)]

                if metadata is None:
                    raise SPyDependencyNotBuilt(self, func, 'component dependency not built')

                func_results = func(self, metadata)

                component_definitions = ItemGroup()
                if func_results is not None:
                    if not isinstance(func_results, list):
                        func_results = [func_results]

                    for func_result in func_results:
                        if isinstance(func_result, _AssetBase):
                            _asset_obj = func_result  # type: _AssetBase
                            if not _common.present(_asset_obj.definition, 'Name'):
                                _asset_obj.definition['Name'] = get_method_friendly_name(func)
                            build_results = _asset_obj.build(metadata)
                            component_definitions.extend(build_results)
                        elif isinstance(func_result, dict):
                            component_definition = func_result  # type: dict
                            _AssetBase._add_asset_metadata(self, component_definition)
                            component_definitions.append(component_definition)

                if self.context.phase == BuildPhase.BUILDING:
                    # We do not want to store the definitions in the cache if we are in the INSTANTIATING phase. In
                    # fact, in that phase it would always be an empty list, which would not be correct to cache.
                    self.context.cache[(self, func.__name__)] = component_definitions
                    self.context.add_results(component_definitions)

                self.context.at_least_one_thing_built_somewhere = True

                return component_definitions

            _AssetBase._set_wrapper_attrs(component_wrapper, MethodType.COMPONENT, func)

            return component_wrapper

        return component_decorator

    # noinspection PyPep8Naming
    @classmethod
    def Display(cls, analysis=None):
        def display_decorator(func):
            # noinspection PyUnusedLocal
            def display_wrapper(self, metadata=None):
                if (self, func.__name__) in self.context.cache:
                    # We've already built this
                    return self.context.cache[(self, func.__name__)]

                if ('analysis', analysis) not in self.context.workbooks:
                    raise WorkbookNotBuilt('analysis', analysis)

                worksheet_object = self.context.workbooks[('analysis', analysis)]  # type: AnalysisWorksheet

                workstep_object = func(self, metadata, worksheet_object)

                self.context.cache[(self, func.__name__)] = workstep_object
                self.context.at_least_one_thing_built_somewhere = True

                return workstep_object

            _AssetBase._set_wrapper_attrs(display_wrapper, MethodType.DISPLAY, func)

            return display_wrapper

        return display_decorator

    # noinspection PyPep8Naming
    @classmethod
    def DateRange(cls):
        def date_range_decorator(func):
            # noinspection PyUnusedLocal
            def date_range_wrapper(self, metadata=None):
                if (self, func.__name__) in self.context.cache:
                    # We've already built this
                    return self.context.cache[(self, func.__name__)]

                date_range_spec = func(self, metadata)
                date_range_object = DateRange(date_range_spec, None)

                if 'Name' not in date_range_spec:
                    date_range_spec['Name'] = get_method_friendly_name(func)

                self.context.cache[(self, func.__name__)] = date_range_object
                self.context.at_least_one_thing_built_somewhere = True

                return date_range_object

            _AssetBase._set_wrapper_attrs(date_range_wrapper, MethodType.DATE_RANGE, func)

            return date_range_wrapper

        return date_range_decorator

    # noinspection PyPep8Naming
    @classmethod
    def Document(cls, topic=None):
        def document_decorator(func):
            # noinspection PyUnusedLocal
            def document_wrapper(self, metadata=None):
                # type: (_AssetBase, pd.DataFrame) -> TopicDocument

                if (self, func.__name__) in self.context.cache:
                    # We've already built this
                    return self.context.cache[(self, func.__name__)]

                if ('topic', topic) not in self.context.workbooks:
                    raise WorkbookNotBuilt('topic', topic)

                topic_object = self.context.workbooks[('topic', topic)]  # type: Topic

                document_object = func(self, metadata, topic_object)

                self.context.cache[(self, func.__name__)] = document_object
                self.context.at_least_one_thing_built_somewhere = True

                return document_object

            _AssetBase._set_wrapper_attrs(document_wrapper, MethodType.DOCUMENT, func)

            return document_wrapper

        return document_decorator

    # noinspection PyPep8Naming
    @classmethod
    def Plot(cls, image_format):
        def plot_decorator(func):
            # noinspection PyUnusedLocal
            def plot_wrapper(self, metadata=None):
                if (self, func.__name__) in self.context.cache:
                    # We've already built this
                    return self.context.cache[(self, func.__name__)]

                def _plot_function(date_range):
                    return func(self, metadata, date_range)

                plot_render_info = PlotRenderInfo(image_format, _plot_function)
                self.context.cache[(self, func.__name__)] = plot_render_info
                self.context.at_least_one_thing_built_somewhere = True

                return plot_render_info

            _AssetBase._set_wrapper_attrs(plot_wrapper, MethodType.PLOT, func)

            return plot_wrapper

        return plot_decorator

    # noinspection PyPep8Naming
    @classmethod
    def Requirement(cls):
        def requirement_decorator(func):
            def requirement_wrapper(self, metadata=None):
                # type: (Asset, pd.DataFrame) -> Optional[dict, tuple]
                if (self, func.__name__) in self.context.cache:
                    return self.context.cache[(self, func.__name__)]

                if metadata is None and self.context.mode != BuildMode.BROCHURE:
                    raise SPyDependencyNotBuilt(self, func, 'requirement dependency not built')

                func_results = func(self, metadata)

                if isinstance(func_results, type):
                    if not issubclass(func_results, Asset):
                        raise ValueError(f'Requirement function returns non-Asset type: {func_results}')

                    if self.context.mode != BuildMode.BROCHURE:
                        return None

                    func_results = {
                        'Type': 'Template',
                        'Name': func_results.__name__.replace('_', ' '),
                        'Description': func_results.__doc__
                    }

                if self.context.mode == BuildMode.BROCHURE:
                    return func_results

                if not _common.present(func_results, 'Name'):
                    func_results['Name'] = get_method_friendly_name(func)

                asset_metadata: pd.DataFrame = metadata[(metadata['Build Path'] == self['Path']) &
                                                        (metadata['Build Asset'] == self['Asset']) &
                                                        (metadata['Type'] == 'Asset')]

                def _handle_optional():
                    if _common.get(func_results, 'Optional'):
                        self.context.cache[(self, func.__name__)] = None
                        return None
                    else:
                        raise SPyAssetMemberError(
                            self, func,
                            f'Requirement "{func_results["Name"]}" not found (required by "{self.__class__.__name__}")')

                if func_results['Type'] == 'Property':
                    if len(asset_metadata) != 0 and func_results['Name'] in asset_metadata.columns:
                        property_name = func_results['Name']
                        property_value = asset_metadata.iloc[0][func_results['Name']]

                        if pd.isna(property_value):
                            return _handle_optional()

                        self.context.cache[(self, func.__name__)] = property_value
                        self.definition[property_name] = property_value
                        return property_value

                    else:
                        return _handle_optional()

                requirement_metadata = metadata[(metadata['Build Path'] == self['Path']) &
                                                (metadata['Build Asset'] == self['Asset']) &
                                                (metadata['Name'] == func_results['Name'])]

                if len(requirement_metadata) == 0:
                    if _common.get(func_results, 'Type') == 'Scalar' and \
                            len(asset_metadata) != 0 and func_results['Name'] in asset_metadata.columns:

                        # This is the case where the user is satisfying a Scalar requirement by providing a column in
                        # the metadata DataFrame that specifies the Scalar value directly (as opposed to referring to
                        # an already-established Scalar item in Seeq).

                        scalar_name = func_results['Name']
                        scalar_value = asset_metadata.iloc[0][func_results['Name']]

                        if pd.isna(scalar_value):
                            return _handle_optional()

                        if _common.get(func_results, 'Unit Of Measure Family') == 'string':
                            scalar_value = _common.string_to_formula_literal(str(scalar_value))

                        requirement_metadata = {
                            'Name': scalar_name,
                            'Type': 'Scalar',
                            'Formula': scalar_value
                        }

                    else:
                        return _handle_optional()

                elif len(requirement_metadata) > 1:
                    raise SPyAssetMemberError(
                        self, func,
                        f'Duplicates found for "{func_results["Name"]}" (required by "{self.__class__.__name__}"):\n'
                        f'{requirement_metadata}')

                requirement_definition = _AssetBase._build_attribute_definition(self, func, requirement_metadata)

                requirement_definition['Requirement'] = True
                requirement_definition['Build Result'] = BuildPhase.SUCCESS

                self.context.cache[(self, func.__name__)] = requirement_definition
                self.context.add_results(requirement_definition)
                self.context.at_least_one_thing_built_somewhere = True

                return requirement_definition

            _AssetBase._set_wrapper_attrs(requirement_wrapper, MethodType.REQUIREMENT, func)

            return requirement_wrapper

        return requirement_decorator


class Asset(_AssetBase):
    """
    A class derived from Asset can have @Asset.Attribute and @Asset.Component decorated functions that are executed
    as part of the call to build() which returns a list of definition dicts for the asset.
    """

    # This class does not differ from _AssetBase... but we separate it from Mixin so that
    # isinstance(mixin_object, Asset) doesn't return True and make things confusing.
    pass


class Mixin(_AssetBase):
    """
    A Mixin is nearly identical to an Asset, but it adds attributes/components to an otherwise-defined Asset. The
    definition argument that is passed into the constructor should be the definition of the otherwise-defined Asset.

    This allows asset tree designers to add a set of "special" attributes to a particular instance of an asset
    without "polluting" the main asset definition with attributes that most of the instances shouldn't have.
    """

    asset: Asset

    def __init__(self, context, definition, parent):
        # Adopt the "parent" asset's definition since this is a Mixin
        super().__init__(context, parent.definition, parent=parent.parent)

        self.asset = parent
        self.definition['Asset Object'] = self
        self.definition['Template'] = self.template_friendly_name


def do_paths_match_criteria(criteria, names):
    return do_lists_match_criteria(_common.path_string_to_list(criteria), _common.path_string_to_list(names))


def do_lists_match_criteria(criteria, names):
    end = max(len(criteria), len(names))
    for i in range(0, end):
        if i > len(criteria) - 1:
            return False

        criterion = criteria[i]
        if criterion == '**':
            remaining_criteria = criteria[i + 1:]
            for j in range(i, len(names) + 1):
                remaining_names = names[j:]
                if do_lists_match_criteria(remaining_criteria, remaining_names):
                    return True

        if i > len(names) - 1:
            return False

        name = names[i]
        if not _common.does_query_fragment_match(criterion, name, contains=False):
            return False

    return True


class ItemGroup(list):
    """
    Represents a list of asset tree items that can be rolled up into a calculation
    """

    def __init__(self, _list=None):
        super().__init__(list(ItemGroup.flatten(_list) if _list else list()))

    @staticmethod
    def flatten(l):
        for el in l:
            if isinstance(el, list):
                yield from ItemGroup.flatten(el)
            else:
                yield el

    def pick(self, criteria: dict) -> ItemGroup:
        """
        Filters items based on given criteria and returns them in a new ItemGroup

        Parameters
        ----------
        criteria : dict
            An dictionary mapping property names to property values

        Examples
        --------
        Return an ItemGroup composed of all items in `item_group` that have type Signal:

        >>> item_group.pick({'Type': 'Signal'})
        """
        picked = ItemGroup()

        if not isinstance(criteria, dict):
            raise SPyValueError('pick(criteria) argument must be a dict')

        for definition in self:
            if definition is None:
                continue

            good = True
            for prop, criterion in criteria.items():
                if prop not in definition:
                    good = False
                    break

                val = definition[prop]

                # If they didn't specify Calculated vs Stored, match on both
                if criterion in ['Signal', 'Condition', 'Scalar']:
                    criterion = f'*{criterion}'

                if prop == 'Path':
                    if not do_paths_match_criteria(criterion, val):
                        good = False
                        break
                elif not _common.does_query_fragment_match(criterion, val, contains=False):
                    good = False
                    break

            if good:
                picked.append(definition)

        return picked

    def assets(self):
        picked = list()
        for definition in self:
            if definition is None:
                continue

            if 'Asset Object' in definition and definition['Asset Object'] not in picked:
                picked.append(definition['Asset Object'])

        return picked

    def as_parameters(self, *, prefix='p'):
        parameters = dict()
        zero_padding = int(math.log10(len(self))) + 1 if len(self) > 0 else 1
        for i in range(len(self)):
            definition = self[i]

            if definition is None:
                # We skip over None here, because it may be due to an optional requirement. This relieves the user
                # from having to always account for "holes" in the tree whenever doing roll-ups
                continue

            string_spec = '$%s{0:0%dd}' % (prefix, zero_padding)
            parameters[string_spec.format(i)] = definition

        return parameters

    def roll_up(self, statistic):
        """
        Returns a definition for a roll-up calculation that takes all items
        in this group as inputs.

        Examples
        --------
        Calculate the average of all items in item_group:

        >>> item_group.roll_up('Average')
        """
        parameters = self.as_parameters()

        if len(parameters) > 0:
            types = {p['Type'] for p in parameters.values()}
            if len(types) == 0:
                return None

            if len(types) > 1:
                debug_string = '\n'.join(
                    [_common.safe_json_dumps(p) for p in parameters.values()])
                raise SPyRuntimeError('Cannot compute statistics across different types of items:\n%s' % debug_string)

            _type = types.pop().replace('Stored', '').replace('Calculated', '')
            fns = {fn.statistic: fn for fn in _common.ROLL_UP_FUNCTIONS if fn.input_type == _type}

            if statistic.lower() not in fns:
                raise SPyValueError('Statistic "%s" not recognized for type %s. Valid statistics:\n%s' %
                                    (statistic, _type, '\n'.join([string.capwords(s) for s in fns.keys()])))
        else:
            fns = {fn.statistic: fn for fn in _common.ROLL_UP_FUNCTIONS}
            if statistic.lower() not in fns:
                raise SPyValueError('Statistic "%s" not recognized. Valid statistics:\n%s' %
                                    (statistic, '\n'.join([string.capwords(s) for s in fns.keys()])))

        fn = fns[statistic.lower()]  # type: _common.RollUpFunction
        formula = fn.generate_formula(parameters)

        return {
            'Type': fn.output_type,
            'Formula': formula,
            'Formula Parameters': parameters
        }


class SPyAssetMemberError(SPyException):
    asset: _AssetBase
    func: Optional[Callable]
    message: str

    def __init__(self, asset: _AssetBase, func: Optional[Callable], message: str):
        self.asset = asset
        self.func = func
        self.message = message

    def __str__(self):
        return f'"{SPyAssetMemberError.get_friendly_path_for_error(self.asset, self.func)}":' \
               f' {self.message}'

    @staticmethod
    def get_friendly_path_for_error(asset: _AssetBase, func: Optional[Callable]):
        full_path = asset.definition.get('Asset', '')
        if _common.present(asset.definition, 'Name') and _common.get(asset.definition, 'Type') != 'Asset':
            full_path += ' >> ' + asset.definition['Name']
        if _common.present(asset.definition, 'Path'):
            full_path = asset.definition['Path'] + ' >> ' + full_path

        if func is not None:
            if hasattr(func, FRIENDLY_NAME_ATTR):
                full_path += ' >> ' + getattr(func, FRIENDLY_NAME_ATTR)
            else:
                full_path += ' >> ' + get_method_friendly_name(func)

        full_path += f' [on {asset.__class__.__name__} class]'

        return full_path


class SPyDependencyNotBuilt(SPyAssetMemberError):
    def __str__(self):
        return f'"{SPyAssetMemberError.get_friendly_path_for_error(self.asset, self.func)}"' \
               f' {self.message}'



from __future__ import annotations

from typing import List, Optional, Union

import pandas as pd

from seeq import spy
from seeq.spy import _common
from seeq.spy import _login
from seeq.spy._errors import *
from seeq.spy._session import Session
from seeq.spy._status import Status
from seeq.spy.workbooks._workbook import Workbook


def pull(workbooks_df: Union[pd.DataFrame, str], *, include_referenced_workbooks: bool = True,
         include_inventory: bool = True, include_images: bool = True, include_rendered_content: bool = False,
         specific_worksheet_ids: Optional[List[str]] = None, errors: str = 'raise', quiet: bool = False,
         status: Optional[Status] = None, session: Optional[Session] = None) -> Optional[List[Workbook]]:
    """
    Pulls the definitions for each workbook specified by workbooks_df into
    memory as a list of Workbook items.

    Parameters
    ----------
    workbooks_df : {str, pd.DataFrame}
        A DataFrame containing 'ID', 'Type' and 'Workbook Type' columns that
        can be used to identify the workbooks to pull. This is usually created
        via a call to spy.workbooks.search(). Alternatively, you can supply URL
        of a Seeq Workbench worksheet as a str.

    include_referenced_workbooks : bool, default True
        If True, Analyses that are depended upon by Topics will be
        automatically included in the resulting list even if they were not part
        of the workbooks_df DataFrame.

    include_inventory : bool, default True
        If True, all items that are referenced in worksheets, journals, topics
        etc are included in the Workbook object's "inventory", along with
        anything that is scoped to the workbook.

    include_images : bool, default True
        If True, downloads all static images (not including embedded content --
        see include_rendered_content for that).

    include_rendered_content : bool, default False
        If True, any Organizer Topics pulled will include rendered content
        images, which will cause spy.workbooks.save() to include a folder for
        each Topic Document with its HTML and images such that it can be loaded
        and viewed in a browser.

    specific_worksheet_ids : List[str], default None
        If supplied, only the worksheets with IDs specified in the supplied
        list will be pulled. This should be used when it would otherwise take
        too long to pull all worksheets and you're only interested in a small
        subset. Be careful not to push the resulting workbook(s) back because
        the other worksheets will get archived.

    errors : {'raise', 'catalog'}, default 'raise'
        If 'raise', any errors encountered will cause an exception. If
        'catalog', errors will be added to a 'Result' column in the status.df
        DataFrame (errors='catalog' must be combined with
        status=<Status object>).

    quiet : bool
        If True, suppresses progress output. Note that when status is
        provided, the quiet setting of the Status object that is passed
        in takes precedent.

    status : spy.Status, optional
        If specified, the supplied Status object will be updated as the command
        progresses. It gets filled in with the same information you would see
        in Jupyter in the blue/green/red table below your code while the
        command is executed. The table itself is accessible as a DataFrame via
        the status.df property.

    session : spy.Session, optional
        If supplied, the Session object (and its Options) will be used to
        store the login session state. This is useful to log in to different
        Seeq servers at the same time or with different credentials.

    """
    _common.validate_argument_types([
        (workbooks_df, 'workbooks_df', (pd.DataFrame, str)),
        (include_referenced_workbooks, 'include_referenced_workbooks', bool),
        (include_inventory, 'include_inventory', bool),
        (include_images, 'include_images', bool),
        (include_rendered_content, 'include_rendered_content', bool),
        (errors, 'errors', str),
        (quiet, 'quiet', bool),
        (status, 'status', Status),
        (session, 'session', Session)
    ])

    status = Status.validate(status, quiet)
    session = Session.validate(session)
    _login.validate_login(session, status)

    _common.validate_errors_arg(errors)
    _common.validate_unique_dataframe_index(workbooks_df, 'workbooks_df')

    if isinstance(workbooks_df, str):
        # If workbooks_df is a URL, get the actual items from the worksheet and overwrite workbooks_df as a DataFrame
        url = workbooks_df
        workbooks_df = spy.workbooks.search({'ID': spy.utils.get_workbook_id_from_url(url)},
                                            quiet=True, errors=errors,
                                            status=status.create_inner('Workbook URL Search'), session=session)
        if len(workbooks_df) == 0:
            raise SPyValueError(f'"{url}" is not a workbook URL')

    if len(workbooks_df) == 0:
        status.update('workbooks_df is empty', Status.SUCCESS)
        return list()

    for required_column in ['ID', 'Type', 'Workbook Type']:
        if required_column not in workbooks_df.columns:
            raise SPyValueError('"%s" column must be included in workbooks_df' % required_column)

    status_columns = list()

    for col in ['ID', 'Name', 'Workbook Type']:
        if col in workbooks_df:
            status_columns.append(col)

    workbooks_df = workbooks_df[workbooks_df['Type'] != 'Folder']

    status.df = workbooks_df[status_columns].copy().reset_index(drop=True)
    status.df['Count'] = 0
    status.df['Time'] = 0
    status.df['Result'] = 'Queued'
    status_columns.extend(['Count', 'Time', 'Result'])

    status.update('Pulling workbooks', Status.RUNNING)

    workbooks_to_pull = {_common.get(row, 'ID'): {
        'Index': index,
        'Workbook Type': _common.get(row, 'Workbook Type'),
        'Search Folder ID': _common.get(row, 'Search Folder ID'),
        'Worksteps': set()
    } for index, row in workbooks_df.iterrows()}

    # Process Topics first so that we can discover Analyses and Worksteps (of embedded content) to add to our list (
    # assuming include_referenced_workbooks is True). Note that it is possible for Analyses to reference Topics (via
    # a link in a Journal) and they will not be properly included in the output. We have accepted this hole because
    # otherwise if we add a Topic to our list and that Topic refers to a workstep in an already-processed Analysis,
    # we would have to go back and re-pull the Analysis. (Maybe someday we'll do that, but not now. :-) )
    for phase in ['Topic', 'Analysis']:
        while True:
            at_least_one_item_pulled = False
            for item_id, pull_info in workbooks_to_pull.copy().items():
                try:
                    if not isinstance(pull_info, dict):
                        # Already pulled
                        continue

                    if pull_info['Workbook Type'] != phase:
                        continue

                    status.reset_timer()
                    status.current_df_index = pull_info['Index']
                    status.put('Result', 'Pulling')

                    workbook = Workbook.pull(item_id,
                                             extra_workstep_tuples=pull_info['Worksteps'],
                                             specific_worksheet_ids=specific_worksheet_ids,
                                             include_inventory=include_inventory,
                                             include_images=include_images,
                                             errors=errors,
                                             status=status,
                                             session=session)  # type: Workbook

                    if include_rendered_content:
                        status.put('Result', 'Pulling rendered content')
                        workbook.pull_rendered_content(errors=errors, quiet=quiet,
                                                       status=status.create_inner(
                                                           f'Pull Embedded Content {workbook}'),
                                                       session=session)

                    # The "Search Folder ID" is a means by which we can establish "relative paths" like a file
                    # system. The idea is that whatever folder you specified for spy.workbooks.search() is probably
                    # the folder to serve as the "start" and all subfolders become relative to whatever folder is
                    # specified as the "start" during the spy.workbook.push() call. ("start" in this case is very
                    # similar to os.path.relpath()'s "start" argument.) If we didn't have this relative mechanism, it
                    # would be difficult to take a folder full of stuff and duplicate it or put it in a new/different
                    # location.
                    if pull_info['Search Folder ID']:
                        workbook['Search Folder ID'] = pull_info['Search Folder ID']
                    elif len(workbook['Ancestors']) > 0:
                        # If there was no Search Folder ID in the pull_info BUT there are Ancestors, that means this
                        # workbook is being pulled without having done spy.workbook.search() first. If so, then make
                        # the closest ancestor folder its "start", that way when spy.workbooks.push() comes around,
                        # it won't be trying to push any containing folders.
                        workbook['Search Folder ID'] = workbook['Ancestors'][-1]

                    if include_referenced_workbooks:
                        def _add_if_necessary(_workbook_id, _workstep_tuples):
                            if _workbook_id not in workbooks_to_pull:
                                to_add_df = spy.workbooks.search({'ID': _workbook_id},
                                                                 errors=errors,
                                                                 status=status.create_inner(
                                                                     f'Find Workbook {_workbook_id}',
                                                                     quiet=True),
                                                                 session=session)
                                if len(to_add_df) == 0:
                                    # Workbook with this ID not found, probably as a result of a tool making an
                                    # invalid link in a Journal or Topic Document.
                                    return

                                workbook_type_to_add = _common.get(to_add_df.iloc[0], 'Workbook Type')
                                if phase == 'Analysis' and workbook_type_to_add == 'Topic':
                                    # We can't process more Topics once we're in the Analysis phase. (See comment
                                    # above the "for phase" loop.)
                                    return

                                workbooks_to_pull[_workbook_id] = {
                                    'Index': len(status.df),
                                    'Workbook Type': workbook_type_to_add,
                                    'Search Folder ID': pull_info['Search Folder ID'],
                                    'Worksteps': set()
                                }

                                to_add_df['Count'] = 0
                                to_add_df['Time'] = 0
                                to_add_df['Result'] = 'Queued'
                                status.df.loc[len(status.df)] = to_add_df.iloc[0][status_columns]

                            if not isinstance(workbooks_to_pull[_workbook_id], dict):
                                # We've already pulled it, that ship has sailed. This can happen when there are a
                                # web of links between Topics / Journals. We only do a "best effort" to make the
                                # links work.
                                return

                            workbooks_to_pull[_workbook_id]['Worksteps'].update(_workstep_tuples)

                        for workbook_id, workstep_tuples in workbook.referenced_workbooks.items():
                            _add_if_necessary(workbook_id, workstep_tuples)

                        for workbook_id, workstep_tuples in \
                                workbook.find_workbook_links(session, errors, status).items():
                            _add_if_necessary(workbook_id, workstep_tuples)

                    workbooks_to_pull[item_id] = workbook
                    at_least_one_item_pulled = True

                    status.put('Time', status.get_timer())
                    status.put('Result', 'Success')

                    if len(workbook.item_pull_errors) > 0:
                        if errors == 'catalog':
                            status.warn(workbook.items_pull_errors_str)
                        else:
                            raise SPyRuntimeError(workbook.items_pull_errors_str)

                except BaseException as e:
                    _common.raise_or_catalog(errors, status, e=e)

            if not at_least_one_item_pulled:
                break

    status.update('Pull successful', Status.SUCCESS)

    return list(workbooks_to_pull.values())



from __future__ import annotations

from typing import List

from seeq.sdk import *
from seeq.spy import _common
from seeq.spy._redaction import safely
from seeq.spy._session import Session
from seeq.spy._status import Status
from seeq.spy.workbooks._item import Item
from seeq.spy.workbooks._user import ItemWithOwnerAndAcl

SHARED = '__Shared__'
CORPORATE = '__Corporate__'
ALL = '__All__'
USERS = '__Users__'
MY_FOLDER = '__My_Folder__'
PUBLIC = '__Public__'

SYNTHETIC_FOLDERS = [MY_FOLDER, SHARED, CORPORATE, ALL, USERS, PUBLIC]


def massage_ancestors(session: Session, item: Item):
    if not session.corporate_folder:
        return

    if len(item.definition['Ancestors']) > 0 and item.definition['Ancestors'][0] == session.corporate_folder.id:
        # Why replace the Corporate folder ID with the CORPORATE string token? The user doesn't want to have to
        # deal with the ID of the corporate folder, especially when importing/exporting workbooks from one server to
        # another.
        item.definition['Ancestors'][0] = CORPORATE


def synthetic_folder_to_content_filter(synthetic_folder):
    # If synthetic_folder is MY_FOLDER, the corresponding content_filter would be 'OWNER', otherwise,  it would be the
    #  synthetic_folder formatted for use by the Folders API
    content_filter = 'OWNER' if synthetic_folder == MY_FOLDER else synthetic_folder.replace('__', '').upper()
    return content_filter


class Folder(ItemWithOwnerAndAcl):
    def _pull(self, session: Session, item_id, errors, status):
        folders_api = FoldersApi(session.client)
        folder_output = safely(lambda: folders_api.get_folder(folder_id=item_id),
                               action_description=f'get Folder {item_id}',
                               errors=errors, status=status)  # type: FolderOutputV1
        if folder_output is None:
            return
        self._pull_owner_and_acl(session, folder_output.owner, errors, status)
        self._pull_ancestors(session, folder_output.ancestors)
        self.provenance = Item.PULL

    def _pull_ancestors(self, session: Session, ancestors: List[ItemPreviewV1]):
        super()._pull_ancestors(session, ancestors)
        massage_ancestors(session, self)

    def _find_by_name(self, session: Session, folder_id, errors: str, status: Status):
        folders_api = FoldersApi(session.client)

        if folder_id and folder_id != _common.PATH_ROOT:
            folders = safely(lambda: folders_api.get_folders(filter='owner',
                                                             folder_id=folder_id,
                                                             limit=10000),
                             action_description=f'get "owner" folders within {folder_id}',
                             additional_errors=[400],
                             errors=errors, status=status)  # type: WorkbenchItemOutputListV1
        else:
            folders = folders_api.get_folders(filter='owner',
                                              limit=10000)  # type: WorkbenchItemOutputListV1

        content_dict = {content.name.lower(): content for content in folders.content}
        if self.name.lower() in content_dict and content_dict[self.name.lower()].type == 'Folder':
            return content_dict[self.name.lower()]

        return None

    def push(self, session: Session, parent_folder_id, datasource_maps, datasource_output, item_map, *, owner=None,
             label=None, access_control=None, errors='raise', status=None):
        items_api = ItemsApi(session.client)
        folders_api = FoldersApi(session.client)
        folder_item = self.find_me(session, label, datasource_output)

        if folder_item is None and self.provenance == Item.CONSTRUCTOR:
            folder_item = self._find_by_name(session, parent_folder_id, errors, status)

        if not folder_item:
            folder_input = FolderInputV1()
            folder_input.name = self['Name']
            if 'Description' in self:
                folder_input.description = self['Description']
            folder_input.owner_id = self.decide_owner(session, datasource_maps, item_map, owner=owner)
            folder_input.parent_folder_id = parent_folder_id if parent_folder_id != _common.PATH_ROOT else None

            folder_output = safely(lambda: folders_api.create_folder(body=folder_input),
                                   action_description=f'create Folder {folder_input.name}',
                                   errors=errors, status=status)  # type: FolderOutputV1
            if folder_output is None:
                return None

            items_api.set_properties(id=folder_output.id, body=[
                ScalarPropertyV1(name='Datasource Class', value=datasource_output.datasource_class),
                ScalarPropertyV1(name='Datasource ID', value=datasource_output.datasource_id),
                ScalarPropertyV1(name='Data ID', value=self._construct_data_id(label))])
        else:
            folder_output = safely(lambda: folders_api.get_folder(folder_id=folder_item.id),
                                   action_description=f'get Folder {folder_item.id}',
                                   errors=errors, status=status)  # type: FolderOutputV1
            if folder_output is None:
                return None

            props = [ScalarPropertyV1(name='Name', value=self['Name'])]
            if 'Description' in self:
                props.append(ScalarPropertyV1(name='Description', value=self['Description']))

            # If the folder happens to be archived, un-archive it. If you're pushing a new copy it seems likely
            # you're intending to revive it.
            props.append(ScalarPropertyV1(name='Archived', value=False))

            items_api.set_properties(id=folder_output.id, body=props)

            owner_id = self.decide_owner(session, datasource_maps, item_map, owner=owner,
                                         current_owner_id=folder_output.owner.id)

            self._push_owner_and_location(session, folder_output, owner_id, parent_folder_id, errors, status)

        item_map[self.id.upper()] = folder_output.id.upper()

        if access_control:
            self._push_acl(session, folder_output.id, datasource_maps, item_map, access_control, errors, status)

        return folder_output



# coding: utf-8

"""
    Seeq REST API

    No description provided (generated by Swagger Codegen https://github.com/swagger-api/swagger-codegen)  # noqa: E501

    OpenAPI spec version: 57.1.1-SNAPSHOT
    
    Generated by: https://github.com/swagger-api/swagger-codegen.git
"""

from __future__ import absolute_import

import sys
import os
import re

# python 2 and python 3 compatibility library
from six import iteritems

from ..configuration import Configuration
from ..api_client import ApiClient
from ..models import *

class WorkbooksApi(object):
    """
    NOTE: This class is auto generated by the swagger code generator program.
    Do not edit the class manually.
    Ref: https://github.com/swagger-api/swagger-codegen
    """

    def __init__(self, api_client=None):
        config = Configuration()
        if api_client:
            self.api_client = api_client
        else:
            if not config.api_client:
                config.api_client = ApiClient()
            self.api_client = config.api_client

    def add_recently_accessed(self, **kwargs):
        """
        Put the specified item at the top of the recently accessed list for the workbook
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please define a `callback` function
        to be invoked when receiving the response.
        >>> def callback_function(response):
        >>>     pprint(response)
        >>>
        >>> thread = api.add_recently_accessed(workbook_id=workbook_id_value, item_id=item_id_value, callback=callback_function)

        :param callback function: The callback function
            for asynchronous request. (optional)
        :param str workbook_id: ID of the workbook (required)
        :param str item_id: Item to be added (required)
        :return: ItemSearchPreviewListV1
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: ItemSearchPreviewListV1
        """
        kwargs['_return_http_data_only'] = True
        if kwargs.get('callback'):
            return self.add_recently_accessed_with_http_info(**kwargs)
        else:
            (data) = self.add_recently_accessed_with_http_info(**kwargs)
            return data

    def add_recently_accessed_with_http_info(self, **kwargs):
        """
        Put the specified item at the top of the recently accessed list for the workbook
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please define a `callback` function
        to be invoked when receiving the response.
        >>> def callback_function(response):
        >>>     pprint(response)
        >>>
        >>> thread = api.add_recently_accessed_with_http_info(workbook_id=workbook_id_value, item_id=item_id_value, callback=callback_function)

        :param callback function: The callback function
            for asynchronous request. (optional)
        :param str workbook_id: ID of the workbook (required)
        :param str item_id: Item to be added (required)
        :return: ItemSearchPreviewListV1
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: ItemSearchPreviewListV1
        """

        all_params = ['workbook_id', 'item_id']
        all_params.append('callback')
        all_params.append('_return_http_data_only')
        all_params.append('_preload_content')
        all_params.append('_request_timeout')

        params = locals()
        for key, val in iteritems(params['kwargs']):
            if key not in all_params:
                raise TypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method add_recently_accessed" % key
                )
            params[key] = val
        del params['kwargs']
        # verify the required parameter 'workbook_id' is set
        if ('workbook_id' not in params) or (params['workbook_id'] is None):
            raise ValueError("Missing the required parameter `workbook_id` when calling `add_recently_accessed`")
        # verify the required parameter 'item_id' is set
        if ('item_id' not in params) or (params['item_id'] is None):
            raise ValueError("Missing the required parameter `item_id` when calling `add_recently_accessed`")


        collection_formats = {}

        path_params = {}
        if 'workbook_id' in params:
            path_params['workbookId'] = params['workbook_id']
        if 'item_id' in params:
            path_params['itemId'] = params['item_id']

        query_params = []

        header_params = {}

        form_params = []
        local_var_files = {}

        body_params = None
        # HTTP header `Accept`
        header_params['Accept'] = self.api_client.\
            select_header_accept(['application/vnd.seeq.v1+json'])

        # HTTP header `Content-Type`
        header_params['Content-Type'] = self.api_client.\
            select_header_content_type(['application/vnd.seeq.v1+json', ])

        # Authentication setting
        auth_settings = ['api_key']

        return self.api_client.call_api('/workbooks/{workbookId}/recently-accessed/{itemId}', 'POST',
                                        path_params,
                                        query_params,
                                        header_params,
                                        body=body_params,
                                        post_params=form_params,
                                        files=local_var_files,
                                        response_type='ItemSearchPreviewListV1',
                                        auth_settings=auth_settings,
                                        callback=params.get('callback'),
                                        _return_http_data_only=params.get('_return_http_data_only'),
                                        _preload_content=params.get('_preload_content', True),
                                        _request_timeout=params.get('_request_timeout'),
                                        collection_formats=collection_formats)

    def archive_workbook(self, **kwargs):
        """
        Archive a workbook by id
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please define a `callback` function
        to be invoked when receiving the response.
        >>> def callback_function(response):
        >>>     pprint(response)
        >>>
        >>> thread = api.archive_workbook(id=id_value, callback=callback_function)

        :param callback function: The callback function
            for asynchronous request. (optional)
        :param str id: ID of the workbook to archive (required)
        :return: ArchiveOutputV1
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: ArchiveOutputV1
        """
        kwargs['_return_http_data_only'] = True
        if kwargs.get('callback'):
            return self.archive_workbook_with_http_info(**kwargs)
        else:
            (data) = self.archive_workbook_with_http_info(**kwargs)
            return data

    def archive_workbook_with_http_info(self, **kwargs):
        """
        Archive a workbook by id
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please define a `callback` function
        to be invoked when receiving the response.
        >>> def callback_function(response):
        >>>     pprint(response)
        >>>
        >>> thread = api.archive_workbook_with_http_info(id=id_value, callback=callback_function)

        :param callback function: The callback function
            for asynchronous request. (optional)
        :param str id: ID of the workbook to archive (required)
        :return: ArchiveOutputV1
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: ArchiveOutputV1
        """

        all_params = ['id']
        all_params.append('callback')
        all_params.append('_return_http_data_only')
        all_params.append('_preload_content')
        all_params.append('_request_timeout')

        params = locals()
        for key, val in iteritems(params['kwargs']):
            if key not in all_params:
                raise TypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method archive_workbook" % key
                )
            params[key] = val
        del params['kwargs']
        # verify the required parameter 'id' is set
        if ('id' not in params) or (params['id'] is None):
            raise ValueError("Missing the required parameter `id` when calling `archive_workbook`")


        collection_formats = {}

        path_params = {}
        if 'id' in params:
            path_params['id'] = params['id']

        query_params = []

        header_params = {}

        form_params = []
        local_var_files = {}

        body_params = None
        # HTTP header `Accept`
        header_params['Accept'] = self.api_client.\
            select_header_accept(['application/vnd.seeq.v1+json'])

        # HTTP header `Content-Type`
        header_params['Content-Type'] = self.api_client.\
            select_header_content_type(['application/vnd.seeq.v1+json', ])

        # Authentication setting
        auth_settings = ['api_key']

        return self.api_client.call_api('/workbooks/{id}', 'DELETE',
                                        path_params,
                                        query_params,
                                        header_params,
                                        body=body_params,
                                        post_params=form_params,
                                        files=local_var_files,
                                        response_type='ArchiveOutputV1',
                                        auth_settings=auth_settings,
                                        callback=params.get('callback'),
                                        _return_http_data_only=params.get('_return_http_data_only'),
                                        _preload_content=params.get('_preload_content', True),
                                        _request_timeout=params.get('_request_timeout'),
                                        collection_formats=collection_formats)

    def archive_worksheet(self, **kwargs):
        """
        Archive a worksheet
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please define a `callback` function
        to be invoked when receiving the response.
        >>> def callback_function(response):
        >>>     pprint(response)
        >>>
        >>> thread = api.archive_worksheet(workbook_id=workbook_id_value, worksheet_id=worksheet_id_value, callback=callback_function)

        :param callback function: The callback function
            for asynchronous request. (optional)
        :param str workbook_id: ID of the workbook on which the worksheet exists (required)
        :param str worksheet_id: ID of the worksheet to archive (required)
        :return: ArchiveOutputV1
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: ArchiveOutputV1
        """
        kwargs['_return_http_data_only'] = True
        if kwargs.get('callback'):
            return self.archive_worksheet_with_http_info(**kwargs)
        else:
            (data) = self.archive_worksheet_with_http_info(**kwargs)
            return data

    def archive_worksheet_with_http_info(self, **kwargs):
        """
        Archive a worksheet
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please define a `callback` function
        to be invoked when receiving the response.
        >>> def callback_function(response):
        >>>     pprint(response)
        >>>
        >>> thread = api.archive_worksheet_with_http_info(workbook_id=workbook_id_value, worksheet_id=worksheet_id_value, callback=callback_function)

        :param callback function: The callback function
            for asynchronous request. (optional)
        :param str workbook_id: ID of the workbook on which the worksheet exists (required)
        :param str worksheet_id: ID of the worksheet to archive (required)
        :return: ArchiveOutputV1
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: ArchiveOutputV1
        """

        all_params = ['workbook_id', 'worksheet_id']
        all_params.append('callback')
        all_params.append('_return_http_data_only')
        all_params.append('_preload_content')
        all_params.append('_request_timeout')

        params = locals()
        for key, val in iteritems(params['kwargs']):
            if key not in all_params:
                raise TypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method archive_worksheet" % key
                )
            params[key] = val
        del params['kwargs']
        # verify the required parameter 'workbook_id' is set
        if ('workbook_id' not in params) or (params['workbook_id'] is None):
            raise ValueError("Missing the required parameter `workbook_id` when calling `archive_worksheet`")
        # verify the required parameter 'worksheet_id' is set
        if ('worksheet_id' not in params) or (params['worksheet_id'] is None):
            raise ValueError("Missing the required parameter `worksheet_id` when calling `archive_worksheet`")


        collection_formats = {}

        path_params = {}
        if 'workbook_id' in params:
            path_params['workbookId'] = params['workbook_id']
        if 'worksheet_id' in params:
            path_params['worksheetId'] = params['worksheet_id']

        query_params = []

        header_params = {}

        form_params = []
        local_var_files = {}

        body_params = None
        # HTTP header `Accept`
        header_params['Accept'] = self.api_client.\
            select_header_accept(['application/vnd.seeq.v1+json'])

        # HTTP header `Content-Type`
        header_params['Content-Type'] = self.api_client.\
            select_header_content_type(['application/vnd.seeq.v1+json', ])

        # Authentication setting
        auth_settings = ['api_key']

        return self.api_client.call_api('/workbooks/{workbookId}/worksheets/{worksheetId}', 'DELETE',
                                        path_params,
                                        query_params,
                                        header_params,
                                        body=body_params,
                                        post_params=form_params,
                                        files=local_var_files,
                                        response_type='ArchiveOutputV1',
                                        auth_settings=auth_settings,
                                        callback=params.get('callback'),
                                        _return_http_data_only=params.get('_return_http_data_only'),
                                        _preload_content=params.get('_preload_content', True),
                                        _request_timeout=params.get('_request_timeout'),
                                        collection_formats=collection_formats)

    def create_workbook(self, **kwargs):
        """
        Create a workbook
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please define a `callback` function
        to be invoked when receiving the response.
        >>> def callback_function(response):
        >>>     pprint(response)
        >>>
        >>> thread = api.create_workbook(body=body_value, callback=callback_function)

        :param callback function: The callback function
            for asynchronous request. (optional)
        :param WorkbookInputV1 body: Workbook information (required)
        :return: WorkbookOutputV1
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: WorkbookOutputV1
        """
        kwargs['_return_http_data_only'] = True
        if kwargs.get('callback'):
            return self.create_workbook_with_http_info(**kwargs)
        else:
            (data) = self.create_workbook_with_http_info(**kwargs)
            return data

    def create_workbook_with_http_info(self, **kwargs):
        """
        Create a workbook
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please define a `callback` function
        to be invoked when receiving the response.
        >>> def callback_function(response):
        >>>     pprint(response)
        >>>
        >>> thread = api.create_workbook_with_http_info(body=body_value, callback=callback_function)

        :param callback function: The callback function
            for asynchronous request. (optional)
        :param WorkbookInputV1 body: Workbook information (required)
        :return: WorkbookOutputV1
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: WorkbookOutputV1
        """

        all_params = ['body']
        all_params.append('callback')
        all_params.append('_return_http_data_only')
        all_params.append('_preload_content')
        all_params.append('_request_timeout')

        params = locals()
        for key, val in iteritems(params['kwargs']):
            if key not in all_params:
                raise TypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method create_workbook" % key
                )
            params[key] = val
        del params['kwargs']
        # verify the required parameter 'body' is set
        if ('body' not in params) or (params['body'] is None):
            raise ValueError("Missing the required parameter `body` when calling `create_workbook`")


        collection_formats = {}

        path_params = {}

        query_params = []

        header_params = {}

        form_params = []
        local_var_files = {}

        body_params = None
        if 'body' in params:
            body_params = params['body']
        # HTTP header `Accept`
        header_params['Accept'] = self.api_client.\
            select_header_accept(['application/vnd.seeq.v1+json'])

        # HTTP header `Content-Type`
        header_params['Content-Type'] = self.api_client.\
            select_header_content_type(['application/vnd.seeq.v1+json', ])

        # Authentication setting
        auth_settings = ['api_key']

        return self.api_client.call_api('/workbooks', 'POST',
                                        path_params,
                                        query_params,
                                        header_params,
                                        body=body_params,
                                        post_params=form_params,
                                        files=local_var_files,
                                        response_type='WorkbookOutputV1',
                                        auth_settings=auth_settings,
                                        callback=params.get('callback'),
                                        _return_http_data_only=params.get('_return_http_data_only'),
                                        _preload_content=params.get('_preload_content', True),
                                        _request_timeout=params.get('_request_timeout'),
                                        collection_formats=collection_formats)

    def create_worksheet(self, **kwargs):
        """
        Create a worksheet
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please define a `callback` function
        to be invoked when receiving the response.
        >>> def callback_function(response):
        >>>     pprint(response)
        >>>
        >>> thread = api.create_worksheet(body=body_value, workbook_id=workbook_id_value, callback=callback_function)

        :param callback function: The callback function
            for asynchronous request. (optional)
        :param WorksheetInputV1 body: Worksheet information (required)
        :param str workbook_id: ID of the workbook on which to create the worksheet (required)
        :return: WorksheetOutputV1
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: WorksheetOutputV1
        """
        kwargs['_return_http_data_only'] = True
        if kwargs.get('callback'):
            return self.create_worksheet_with_http_info(**kwargs)
        else:
            (data) = self.create_worksheet_with_http_info(**kwargs)
            return data

    def create_worksheet_with_http_info(self, **kwargs):
        """
        Create a worksheet
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please define a `callback` function
        to be invoked when receiving the response.
        >>> def callback_function(response):
        >>>     pprint(response)
        >>>
        >>> thread = api.create_worksheet_with_http_info(body=body_value, workbook_id=workbook_id_value, callback=callback_function)

        :param callback function: The callback function
            for asynchronous request. (optional)
        :param WorksheetInputV1 body: Worksheet information (required)
        :param str workbook_id: ID of the workbook on which to create the worksheet (required)
        :return: WorksheetOutputV1
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: WorksheetOutputV1
        """

        all_params = ['body', 'workbook_id']
        all_params.append('callback')
        all_params.append('_return_http_data_only')
        all_params.append('_preload_content')
        all_params.append('_request_timeout')

        params = locals()
        for key, val in iteritems(params['kwargs']):
            if key not in all_params:
                raise TypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method create_worksheet" % key
                )
            params[key] = val
        del params['kwargs']
        # verify the required parameter 'body' is set
        if ('body' not in params) or (params['body'] is None):
            raise ValueError("Missing the required parameter `body` when calling `create_worksheet`")
        # verify the required parameter 'workbook_id' is set
        if ('workbook_id' not in params) or (params['workbook_id'] is None):
            raise ValueError("Missing the required parameter `workbook_id` when calling `create_worksheet`")


        collection_formats = {}

        path_params = {}
        if 'workbook_id' in params:
            path_params['workbookId'] = params['workbook_id']

        query_params = []

        header_params = {}

        form_params = []
        local_var_files = {}

        body_params = None
        if 'body' in params:
            body_params = params['body']
        # HTTP header `Accept`
        header_params['Accept'] = self.api_client.\
            select_header_accept(['application/vnd.seeq.v1+json'])

        # HTTP header `Content-Type`
        header_params['Content-Type'] = self.api_client.\
            select_header_content_type(['application/vnd.seeq.v1+json', ])

        # Authentication setting
        auth_settings = ['api_key']

        return self.api_client.call_api('/workbooks/{workbookId}/worksheets', 'POST',
                                        path_params,
                                        query_params,
                                        header_params,
                                        body=body_params,
                                        post_params=form_params,
                                        files=local_var_files,
                                        response_type='WorksheetOutputV1',
                                        auth_settings=auth_settings,
                                        callback=params.get('callback'),
                                        _return_http_data_only=params.get('_return_http_data_only'),
                                        _preload_content=params.get('_preload_content', True),
                                        _request_timeout=params.get('_request_timeout'),
                                        collection_formats=collection_formats)

    def create_workstep(self, **kwargs):
        """
        Append a new workstep to the worksheet
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please define a `callback` function
        to be invoked when receiving the response.
        >>> def callback_function(response):
        >>>     pprint(response)
        >>>
        >>> thread = api.create_workstep(body=body_value, workbook_id=workbook_id_value, worksheet_id=worksheet_id_value, callback=callback_function)

        :param callback function: The callback function
            for asynchronous request. (optional)
        :param WorkstepInputV1 body: Workstep information (required)
        :param str workbook_id: ID of the workbook on which the worksheet exists (required)
        :param str worksheet_id: ID of the worksheet on which to append the new workstep (required)
        :param bool no_workstep_message: Option to not send a workstep message notifying clients that the current workstep of the worksheet has been updated. Should only be used when multiple worksteps are being pushed in quick succession and a subsequent call will be made to set the current workstep.
        :return: WorkstepOutputV1
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: WorkstepOutputV1
        """
        kwargs['_return_http_data_only'] = True
        if kwargs.get('callback'):
            return self.create_workstep_with_http_info(**kwargs)
        else:
            (data) = self.create_workstep_with_http_info(**kwargs)
            return data

    def create_workstep_with_http_info(self, **kwargs):
        """
        Append a new workstep to the worksheet
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please define a `callback` function
        to be invoked when receiving the response.
        >>> def callback_function(response):
        >>>     pprint(response)
        >>>
        >>> thread = api.create_workstep_with_http_info(body=body_value, workbook_id=workbook_id_value, worksheet_id=worksheet_id_value, callback=callback_function)

        :param callback function: The callback function
            for asynchronous request. (optional)
        :param WorkstepInputV1 body: Workstep information (required)
        :param str workbook_id: ID of the workbook on which the worksheet exists (required)
        :param str worksheet_id: ID of the worksheet on which to append the new workstep (required)
        :param bool no_workstep_message: Option to not send a workstep message notifying clients that the current workstep of the worksheet has been updated. Should only be used when multiple worksteps are being pushed in quick succession and a subsequent call will be made to set the current workstep.
        :return: WorkstepOutputV1
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: WorkstepOutputV1
        """

        all_params = ['body', 'workbook_id', 'worksheet_id', 'no_workstep_message']
        all_params.append('callback')
        all_params.append('_return_http_data_only')
        all_params.append('_preload_content')
        all_params.append('_request_timeout')

        params = locals()
        for key, val in iteritems(params['kwargs']):
            if key not in all_params:
                raise TypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method create_workstep" % key
                )
            params[key] = val
        del params['kwargs']
        # verify the required parameter 'body' is set
        if ('body' not in params) or (params['body'] is None):
            raise ValueError("Missing the required parameter `body` when calling `create_workstep`")
        # verify the required parameter 'workbook_id' is set
        if ('workbook_id' not in params) or (params['workbook_id'] is None):
            raise ValueError("Missing the required parameter `workbook_id` when calling `create_workstep`")
        # verify the required parameter 'worksheet_id' is set
        if ('worksheet_id' not in params) or (params['worksheet_id'] is None):
            raise ValueError("Missing the required parameter `worksheet_id` when calling `create_workstep`")


        collection_formats = {}

        path_params = {}
        if 'workbook_id' in params:
            path_params['workbookId'] = params['workbook_id']
        if 'worksheet_id' in params:
            path_params['worksheetId'] = params['worksheet_id']

        query_params = []
        if 'no_workstep_message' in params:
            query_params.append(('noWorkstepMessage', params['no_workstep_message']))

        header_params = {}

        form_params = []
        local_var_files = {}

        body_params = None
        if 'body' in params:
            body_params = params['body']
        # HTTP header `Accept`
        header_params['Accept'] = self.api_client.\
            select_header_accept(['application/vnd.seeq.v1+json'])

        # HTTP header `Content-Type`
        header_params['Content-Type'] = self.api_client.\
            select_header_content_type(['application/vnd.seeq.v1+json', ])

        # Authentication setting
        auth_settings = ['api_key']

        return self.api_client.call_api('/workbooks/{workbookId}/worksheets/{worksheetId}/worksteps', 'POST',
                                        path_params,
                                        query_params,
                                        header_params,
                                        body=body_params,
                                        post_params=form_params,
                                        files=local_var_files,
                                        response_type='WorkstepOutputV1',
                                        auth_settings=auth_settings,
                                        callback=params.get('callback'),
                                        _return_http_data_only=params.get('_return_http_data_only'),
                                        _preload_content=params.get('_preload_content', True),
                                        _request_timeout=params.get('_request_timeout'),
                                        collection_formats=collection_formats)

    def get_recently_accessed(self, **kwargs):
        """
        Get the recently accessed list for the workbook
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please define a `callback` function
        to be invoked when receiving the response.
        >>> def callback_function(response):
        >>>     pprint(response)
        >>>
        >>> thread = api.get_recently_accessed(workbook_id=workbook_id_value, callback=callback_function)

        :param callback function: The callback function
            for asynchronous request. (optional)
        :param str workbook_id: ID of the workbook (required)
        :return: ItemSearchPreviewListV1
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: ItemSearchPreviewListV1
        """
        kwargs['_return_http_data_only'] = True
        if kwargs.get('callback'):
            return self.get_recently_accessed_with_http_info(**kwargs)
        else:
            (data) = self.get_recently_accessed_with_http_info(**kwargs)
            return data

    def get_recently_accessed_with_http_info(self, **kwargs):
        """
        Get the recently accessed list for the workbook
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please define a `callback` function
        to be invoked when receiving the response.
        >>> def callback_function(response):
        >>>     pprint(response)
        >>>
        >>> thread = api.get_recently_accessed_with_http_info(workbook_id=workbook_id_value, callback=callback_function)

        :param callback function: The callback function
            for asynchronous request. (optional)
        :param str workbook_id: ID of the workbook (required)
        :return: ItemSearchPreviewListV1
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: ItemSearchPreviewListV1
        """

        all_params = ['workbook_id']
        all_params.append('callback')
        all_params.append('_return_http_data_only')
        all_params.append('_preload_content')
        all_params.append('_request_timeout')

        params = locals()
        for key, val in iteritems(params['kwargs']):
            if key not in all_params:
                raise TypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method get_recently_accessed" % key
                )
            params[key] = val
        del params['kwargs']
        # verify the required parameter 'workbook_id' is set
        if ('workbook_id' not in params) or (params['workbook_id'] is None):
            raise ValueError("Missing the required parameter `workbook_id` when calling `get_recently_accessed`")


        collection_formats = {}

        path_params = {}
        if 'workbook_id' in params:
            path_params['workbookId'] = params['workbook_id']

        query_params = []

        header_params = {}

        form_params = []
        local_var_files = {}

        body_params = None
        # HTTP header `Accept`
        header_params['Accept'] = self.api_client.\
            select_header_accept(['application/vnd.seeq.v1+json'])

        # HTTP header `Content-Type`
        header_params['Content-Type'] = self.api_client.\
            select_header_content_type(['application/vnd.seeq.v1+json', ])

        # Authentication setting
        auth_settings = ['api_key']

        return self.api_client.call_api('/workbooks/{workbookId}/recently-accessed', 'GET',
                                        path_params,
                                        query_params,
                                        header_params,
                                        body=body_params,
                                        post_params=form_params,
                                        files=local_var_files,
                                        response_type='ItemSearchPreviewListV1',
                                        auth_settings=auth_settings,
                                        callback=params.get('callback'),
                                        _return_http_data_only=params.get('_return_http_data_only'),
                                        _preload_content=params.get('_preload_content', True),
                                        _request_timeout=params.get('_request_timeout'),
                                        collection_formats=collection_formats)

    def get_referenced_items(self, **kwargs):
        """
        Get items referenced by workstep or a journal
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please define a `callback` function
        to be invoked when receiving the response.
        >>> def callback_function(response):
        >>>     pprint(response)
        >>>
        >>> thread = api.get_referenced_items(worksheet_id=worksheet_id_value, workstep_id=workstep_id_value, callback=callback_function)

        :param callback function: The callback function
            for asynchronous request. (optional)
        :param str worksheet_id: ID of the worksheet associated with journal that references items to return (required)
        :param str workstep_id: ID of the workstep that references items to return (required)
        :return: ReferencedItemsOutputV1
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: ReferencedItemsOutputV1
        """
        kwargs['_return_http_data_only'] = True
        if kwargs.get('callback'):
            return self.get_referenced_items_with_http_info(**kwargs)
        else:
            (data) = self.get_referenced_items_with_http_info(**kwargs)
            return data

    def get_referenced_items_with_http_info(self, **kwargs):
        """
        Get items referenced by workstep or a journal
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please define a `callback` function
        to be invoked when receiving the response.
        >>> def callback_function(response):
        >>>     pprint(response)
        >>>
        >>> thread = api.get_referenced_items_with_http_info(worksheet_id=worksheet_id_value, workstep_id=workstep_id_value, callback=callback_function)

        :param callback function: The callback function
            for asynchronous request. (optional)
        :param str worksheet_id: ID of the worksheet associated with journal that references items to return (required)
        :param str workstep_id: ID of the workstep that references items to return (required)
        :return: ReferencedItemsOutputV1
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: ReferencedItemsOutputV1
        """

        all_params = ['worksheet_id', 'workstep_id']
        all_params.append('callback')
        all_params.append('_return_http_data_only')
        all_params.append('_preload_content')
        all_params.append('_request_timeout')

        params = locals()
        for key, val in iteritems(params['kwargs']):
            if key not in all_params:
                raise TypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method get_referenced_items" % key
                )
            params[key] = val
        del params['kwargs']
        # verify the required parameter 'worksheet_id' is set
        if ('worksheet_id' not in params) or (params['worksheet_id'] is None):
            raise ValueError("Missing the required parameter `worksheet_id` when calling `get_referenced_items`")
        # verify the required parameter 'workstep_id' is set
        if ('workstep_id' not in params) or (params['workstep_id'] is None):
            raise ValueError("Missing the required parameter `workstep_id` when calling `get_referenced_items`")


        collection_formats = {}

        path_params = {}
        if 'worksheet_id' in params:
            path_params['worksheetId'] = params['worksheet_id']
        if 'workstep_id' in params:
            path_params['workstepId'] = params['workstep_id']

        query_params = []

        header_params = {}

        form_params = []
        local_var_files = {}

        body_params = None
        # HTTP header `Accept`
        header_params['Accept'] = self.api_client.\
            select_header_accept(['application/vnd.seeq.v1+json'])

        # HTTP header `Content-Type`
        header_params['Content-Type'] = self.api_client.\
            select_header_content_type(['application/vnd.seeq.v1+json', ])

        # Authentication setting
        auth_settings = ['api_key']

        return self.api_client.call_api('/workbooks/worksheets/{worksheetId}/worksteps/{workstepId}/referenced-items', 'GET',
                                        path_params,
                                        query_params,
                                        header_params,
                                        body=body_params,
                                        post_params=form_params,
                                        files=local_var_files,
                                        response_type='ReferencedItemsOutputV1',
                                        auth_settings=auth_settings,
                                        callback=params.get('callback'),
                                        _return_http_data_only=params.get('_return_http_data_only'),
                                        _preload_content=params.get('_preload_content', True),
                                        _request_timeout=params.get('_request_timeout'),
                                        collection_formats=collection_formats)

    def get_workbook(self, **kwargs):
        """
        Get a workbook
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please define a `callback` function
        to be invoked when receiving the response.
        >>> def callback_function(response):
        >>>     pprint(response)
        >>>
        >>> thread = api.get_workbook(id=id_value, callback=callback_function)

        :param callback function: The callback function
            for asynchronous request. (optional)
        :param str id: ID of the workbook to retrieve (required)
        :return: WorkbookOutputV1
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: WorkbookOutputV1
        """
        kwargs['_return_http_data_only'] = True
        if kwargs.get('callback'):
            return self.get_workbook_with_http_info(**kwargs)
        else:
            (data) = self.get_workbook_with_http_info(**kwargs)
            return data

    def get_workbook_with_http_info(self, **kwargs):
        """
        Get a workbook
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please define a `callback` function
        to be invoked when receiving the response.
        >>> def callback_function(response):
        >>>     pprint(response)
        >>>
        >>> thread = api.get_workbook_with_http_info(id=id_value, callback=callback_function)

        :param callback function: The callback function
            for asynchronous request. (optional)
        :param str id: ID of the workbook to retrieve (required)
        :return: WorkbookOutputV1
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: WorkbookOutputV1
        """

        all_params = ['id']
        all_params.append('callback')
        all_params.append('_return_http_data_only')
        all_params.append('_preload_content')
        all_params.append('_request_timeout')

        params = locals()
        for key, val in iteritems(params['kwargs']):
            if key not in all_params:
                raise TypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method get_workbook" % key
                )
            params[key] = val
        del params['kwargs']
        # verify the required parameter 'id' is set
        if ('id' not in params) or (params['id'] is None):
            raise ValueError("Missing the required parameter `id` when calling `get_workbook`")


        collection_formats = {}

        path_params = {}
        if 'id' in params:
            path_params['id'] = params['id']

        query_params = []

        header_params = {}

        form_params = []
        local_var_files = {}

        body_params = None
        # HTTP header `Accept`
        header_params['Accept'] = self.api_client.\
            select_header_accept(['application/vnd.seeq.v1+json'])

        # HTTP header `Content-Type`
        header_params['Content-Type'] = self.api_client.\
            select_header_content_type(['application/vnd.seeq.v1+json', ])

        # Authentication setting
        auth_settings = ['api_key']

        return self.api_client.call_api('/workbooks/{id}', 'GET',
                                        path_params,
                                        query_params,
                                        header_params,
                                        body=body_params,
                                        post_params=form_params,
                                        files=local_var_files,
                                        response_type='WorkbookOutputV1',
                                        auth_settings=auth_settings,
                                        callback=params.get('callback'),
                                        _return_http_data_only=params.get('_return_http_data_only'),
                                        _preload_content=params.get('_preload_content', True),
                                        _request_timeout=params.get('_request_timeout'),
                                        collection_formats=collection_formats)

    def get_workbooks(self, **kwargs):
        """
        Get a collection of workbooks
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please define a `callback` function
        to be invoked when receiving the response.
        >>> def callback_function(response):
        >>>     pprint(response)
        >>>
        >>> thread = api.get_workbooks(callback=callback_function)

        :param callback function: The callback function
            for asynchronous request. (optional)
        :param str user_id: The user ID to return workbooks for, defaults to the current user if not specified. Only an admin user is allowed to specify a user ID.
        :param bool is_archived: True to filter results to only workbooks that have been archived, false to filter results to workbooks that are not archived.
        :param str sort_order: A field by which to order the workbooks followed by a space and 'asc' or 'desc'. Field name can be one of: createdAt, updatedAt, name
        :param int offset: The pagination offset, the index of the first collection item that will be returned in this page of results
        :param int limit: The pagination limit, the total number of collection items that will be returned in this page of results
        :return: WorkbookOutputListV1
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: WorkbookOutputListV1
        """
        kwargs['_return_http_data_only'] = True
        if kwargs.get('callback'):
            return self.get_workbooks_with_http_info(**kwargs)
        else:
            (data) = self.get_workbooks_with_http_info(**kwargs)
            return data

    def get_workbooks_with_http_info(self, **kwargs):
        """
        Get a collection of workbooks
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please define a `callback` function
        to be invoked when receiving the response.
        >>> def callback_function(response):
        >>>     pprint(response)
        >>>
        >>> thread = api.get_workbooks_with_http_info(callback=callback_function)

        :param callback function: The callback function
            for asynchronous request. (optional)
        :param str user_id: The user ID to return workbooks for, defaults to the current user if not specified. Only an admin user is allowed to specify a user ID.
        :param bool is_archived: True to filter results to only workbooks that have been archived, false to filter results to workbooks that are not archived.
        :param str sort_order: A field by which to order the workbooks followed by a space and 'asc' or 'desc'. Field name can be one of: createdAt, updatedAt, name
        :param int offset: The pagination offset, the index of the first collection item that will be returned in this page of results
        :param int limit: The pagination limit, the total number of collection items that will be returned in this page of results
        :return: WorkbookOutputListV1
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: WorkbookOutputListV1
        """

        all_params = ['user_id', 'is_archived', 'sort_order', 'offset', 'limit']
        all_params.append('callback')
        all_params.append('_return_http_data_only')
        all_params.append('_preload_content')
        all_params.append('_request_timeout')

        params = locals()
        for key, val in iteritems(params['kwargs']):
            if key not in all_params:
                raise TypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method get_workbooks" % key
                )
            params[key] = val
        del params['kwargs']


        collection_formats = {}

        path_params = {}

        query_params = []
        if 'user_id' in params:
            query_params.append(('userId', params['user_id']))
        if 'is_archived' in params:
            query_params.append(('isArchived', params['is_archived']))
        if 'sort_order' in params:
            query_params.append(('sortOrder', params['sort_order']))
        if 'offset' in params:
            query_params.append(('offset', params['offset']))
        if 'limit' in params:
            query_params.append(('limit', params['limit']))

        header_params = {}

        form_params = []
        local_var_files = {}

        body_params = None
        # HTTP header `Accept`
        header_params['Accept'] = self.api_client.\
            select_header_accept(['application/vnd.seeq.v1+json'])

        # HTTP header `Content-Type`
        header_params['Content-Type'] = self.api_client.\
            select_header_content_type(['application/vnd.seeq.v1+json', ])

        # Authentication setting
        auth_settings = ['api_key']

        return self.api_client.call_api('/workbooks', 'GET',
                                        path_params,
                                        query_params,
                                        header_params,
                                        body=body_params,
                                        post_params=form_params,
                                        files=local_var_files,
                                        response_type='WorkbookOutputListV1',
                                        auth_settings=auth_settings,
                                        callback=params.get('callback'),
                                        _return_http_data_only=params.get('_return_http_data_only'),
                                        _preload_content=params.get('_preload_content', True),
                                        _request_timeout=params.get('_request_timeout'),
                                        collection_formats=collection_formats)

    def get_worksheet(self, **kwargs):
        """
        Get a worksheet
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please define a `callback` function
        to be invoked when receiving the response.
        >>> def callback_function(response):
        >>>     pprint(response)
        >>>
        >>> thread = api.get_worksheet(workbook_id=workbook_id_value, worksheet_id=worksheet_id_value, callback=callback_function)

        :param callback function: The callback function
            for asynchronous request. (optional)
        :param str workbook_id: ID of the workbook for which to retrieve the worksheets (required)
        :param str worksheet_id: ID of the worksheet to retrieve (required)
        :return: WorksheetOutputV1
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: WorksheetOutputV1
        """
        kwargs['_return_http_data_only'] = True
        if kwargs.get('callback'):
            return self.get_worksheet_with_http_info(**kwargs)
        else:
            (data) = self.get_worksheet_with_http_info(**kwargs)
            return data

    def get_worksheet_with_http_info(self, **kwargs):
        """
        Get a worksheet
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please define a `callback` function
        to be invoked when receiving the response.
        >>> def callback_function(response):
        >>>     pprint(response)
        >>>
        >>> thread = api.get_worksheet_with_http_info(workbook_id=workbook_id_value, worksheet_id=worksheet_id_value, callback=callback_function)

        :param callback function: The callback function
            for asynchronous request. (optional)
        :param str workbook_id: ID of the workbook for which to retrieve the worksheets (required)
        :param str worksheet_id: ID of the worksheet to retrieve (required)
        :return: WorksheetOutputV1
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: WorksheetOutputV1
        """

        all_params = ['workbook_id', 'worksheet_id']
        all_params.append('callback')
        all_params.append('_return_http_data_only')
        all_params.append('_preload_content')
        all_params.append('_request_timeout')

        params = locals()
        for key, val in iteritems(params['kwargs']):
            if key not in all_params:
                raise TypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method get_worksheet" % key
                )
            params[key] = val
        del params['kwargs']
        # verify the required parameter 'workbook_id' is set
        if ('workbook_id' not in params) or (params['workbook_id'] is None):
            raise ValueError("Missing the required parameter `workbook_id` when calling `get_worksheet`")
        # verify the required parameter 'worksheet_id' is set
        if ('worksheet_id' not in params) or (params['worksheet_id'] is None):
            raise ValueError("Missing the required parameter `worksheet_id` when calling `get_worksheet`")


        collection_formats = {}

        path_params = {}
        if 'workbook_id' in params:
            path_params['workbookId'] = params['workbook_id']
        if 'worksheet_id' in params:
            path_params['worksheetId'] = params['worksheet_id']

        query_params = []

        header_params = {}

        form_params = []
        local_var_files = {}

        body_params = None
        # HTTP header `Accept`
        header_params['Accept'] = self.api_client.\
            select_header_accept(['application/vnd.seeq.v1+json'])

        # HTTP header `Content-Type`
        header_params['Content-Type'] = self.api_client.\
            select_header_content_type(['application/vnd.seeq.v1+json', ])

        # Authentication setting
        auth_settings = ['api_key']

        return self.api_client.call_api('/workbooks/{workbookId}/worksheets/{worksheetId}', 'GET',
                                        path_params,
                                        query_params,
                                        header_params,
                                        body=body_params,
                                        post_params=form_params,
                                        files=local_var_files,
                                        response_type='WorksheetOutputV1',
                                        auth_settings=auth_settings,
                                        callback=params.get('callback'),
                                        _return_http_data_only=params.get('_return_http_data_only'),
                                        _preload_content=params.get('_preload_content', True),
                                        _request_timeout=params.get('_request_timeout'),
                                        collection_formats=collection_formats)

    def get_worksheet_usages(self, **kwargs):
        """
        Get a list of reports that reference a given worksheet
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please define a `callback` function
        to be invoked when receiving the response.
        >>> def callback_function(response):
        >>>     pprint(response)
        >>>
        >>> thread = api.get_worksheet_usages(workbook_id=workbook_id_value, worksheet_id=worksheet_id_value, callback=callback_function)

        :param callback function: The callback function
            for asynchronous request. (optional)
        :param str workbook_id: The ID of the workbook for containing the worksheet (required)
        :param str worksheet_id: The ID of the worksheet (required)
        :param int offset: The pagination offset, the index of the first collection item that will be returned in this page of results
        :param int limit: The pagination limit, the total number of collection items that will be returned in this page of results
        :return: AnnotationListOutputV1
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: AnnotationListOutputV1
        """
        kwargs['_return_http_data_only'] = True
        if kwargs.get('callback'):
            return self.get_worksheet_usages_with_http_info(**kwargs)
        else:
            (data) = self.get_worksheet_usages_with_http_info(**kwargs)
            return data

    def get_worksheet_usages_with_http_info(self, **kwargs):
        """
        Get a list of reports that reference a given worksheet
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please define a `callback` function
        to be invoked when receiving the response.
        >>> def callback_function(response):
        >>>     pprint(response)
        >>>
        >>> thread = api.get_worksheet_usages_with_http_info(workbook_id=workbook_id_value, worksheet_id=worksheet_id_value, callback=callback_function)

        :param callback function: The callback function
            for asynchronous request. (optional)
        :param str workbook_id: The ID of the workbook for containing the worksheet (required)
        :param str worksheet_id: The ID of the worksheet (required)
        :param int offset: The pagination offset, the index of the first collection item that will be returned in this page of results
        :param int limit: The pagination limit, the total number of collection items that will be returned in this page of results
        :return: AnnotationListOutputV1
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: AnnotationListOutputV1
        """

        all_params = ['workbook_id', 'worksheet_id', 'offset', 'limit']
        all_params.append('callback')
        all_params.append('_return_http_data_only')
        all_params.append('_preload_content')
        all_params.append('_request_timeout')

        params = locals()
        for key, val in iteritems(params['kwargs']):
            if key not in all_params:
                raise TypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method get_worksheet_usages" % key
                )
            params[key] = val
        del params['kwargs']
        # verify the required parameter 'workbook_id' is set
        if ('workbook_id' not in params) or (params['workbook_id'] is None):
            raise ValueError("Missing the required parameter `workbook_id` when calling `get_worksheet_usages`")
        # verify the required parameter 'worksheet_id' is set
        if ('worksheet_id' not in params) or (params['worksheet_id'] is None):
            raise ValueError("Missing the required parameter `worksheet_id` when calling `get_worksheet_usages`")


        collection_formats = {}

        path_params = {}
        if 'workbook_id' in params:
            path_params['workbookId'] = params['workbook_id']
        if 'worksheet_id' in params:
            path_params['worksheetId'] = params['worksheet_id']

        query_params = []
        if 'offset' in params:
            query_params.append(('offset', params['offset']))
        if 'limit' in params:
            query_params.append(('limit', params['limit']))

        header_params = {}

        form_params = []
        local_var_files = {}

        body_params = None
        # HTTP header `Accept`
        header_params['Accept'] = self.api_client.\
            select_header_accept(['application/vnd.seeq.v1+json'])

        # HTTP header `Content-Type`
        header_params['Content-Type'] = self.api_client.\
            select_header_content_type(['application/vnd.seeq.v1+json', ])

        # Authentication setting
        auth_settings = ['api_key']

        return self.api_client.call_api('/workbooks/{workbookId}/worksheets/{worksheetId}/usages', 'GET',
                                        path_params,
                                        query_params,
                                        header_params,
                                        body=body_params,
                                        post_params=form_params,
                                        files=local_var_files,
                                        response_type='AnnotationListOutputV1',
                                        auth_settings=auth_settings,
                                        callback=params.get('callback'),
                                        _return_http_data_only=params.get('_return_http_data_only'),
                                        _preload_content=params.get('_preload_content', True),
                                        _request_timeout=params.get('_request_timeout'),
                                        collection_formats=collection_formats)

    def get_worksheets(self, **kwargs):
        """
        Get a collection of worksheets
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please define a `callback` function
        to be invoked when receiving the response.
        >>> def callback_function(response):
        >>>     pprint(response)
        >>>
        >>> thread = api.get_worksheets(workbook_id=workbook_id_value, callback=callback_function)

        :param callback function: The callback function
            for asynchronous request. (optional)
        :param str workbook_id: The ID of the workbook for which to retrieve the worksheets (required)
        :param int offset: The pagination offset, the index of the first collection item that will be returned in this page of results
        :param int limit: The pagination limit, the total number of collection items that will be returned in this page of results
        :param bool is_archived: Whether to return archived worksheets
        :return: WorksheetOutputListV1
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: WorksheetOutputListV1
        """
        kwargs['_return_http_data_only'] = True
        if kwargs.get('callback'):
            return self.get_worksheets_with_http_info(**kwargs)
        else:
            (data) = self.get_worksheets_with_http_info(**kwargs)
            return data

    def get_worksheets_with_http_info(self, **kwargs):
        """
        Get a collection of worksheets
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please define a `callback` function
        to be invoked when receiving the response.
        >>> def callback_function(response):
        >>>     pprint(response)
        >>>
        >>> thread = api.get_worksheets_with_http_info(workbook_id=workbook_id_value, callback=callback_function)

        :param callback function: The callback function
            for asynchronous request. (optional)
        :param str workbook_id: The ID of the workbook for which to retrieve the worksheets (required)
        :param int offset: The pagination offset, the index of the first collection item that will be returned in this page of results
        :param int limit: The pagination limit, the total number of collection items that will be returned in this page of results
        :param bool is_archived: Whether to return archived worksheets
        :return: WorksheetOutputListV1
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: WorksheetOutputListV1
        """

        all_params = ['workbook_id', 'offset', 'limit', 'is_archived']
        all_params.append('callback')
        all_params.append('_return_http_data_only')
        all_params.append('_preload_content')
        all_params.append('_request_timeout')

        params = locals()
        for key, val in iteritems(params['kwargs']):
            if key not in all_params:
                raise TypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method get_worksheets" % key
                )
            params[key] = val
        del params['kwargs']
        # verify the required parameter 'workbook_id' is set
        if ('workbook_id' not in params) or (params['workbook_id'] is None):
            raise ValueError("Missing the required parameter `workbook_id` when calling `get_worksheets`")


        collection_formats = {}

        path_params = {}
        if 'workbook_id' in params:
            path_params['workbookId'] = params['workbook_id']

        query_params = []
        if 'offset' in params:
            query_params.append(('offset', params['offset']))
        if 'limit' in params:
            query_params.append(('limit', params['limit']))
        if 'is_archived' in params:
            query_params.append(('isArchived', params['is_archived']))

        header_params = {}

        form_params = []
        local_var_files = {}

        body_params = None
        # HTTP header `Accept`
        header_params['Accept'] = self.api_client.\
            select_header_accept(['application/vnd.seeq.v1+json'])

        # HTTP header `Content-Type`
        header_params['Content-Type'] = self.api_client.\
            select_header_content_type(['application/vnd.seeq.v1+json', ])

        # Authentication setting
        auth_settings = ['api_key']

        return self.api_client.call_api('/workbooks/{workbookId}/worksheets', 'GET',
                                        path_params,
                                        query_params,
                                        header_params,
                                        body=body_params,
                                        post_params=form_params,
                                        files=local_var_files,
                                        response_type='WorksheetOutputListV1',
                                        auth_settings=auth_settings,
                                        callback=params.get('callback'),
                                        _return_http_data_only=params.get('_return_http_data_only'),
                                        _preload_content=params.get('_preload_content', True),
                                        _request_timeout=params.get('_request_timeout'),
                                        collection_formats=collection_formats)

    def get_workstep(self, **kwargs):
        """
        Get a workstep
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please define a `callback` function
        to be invoked when receiving the response.
        >>> def callback_function(response):
        >>>     pprint(response)
        >>>
        >>> thread = api.get_workstep(workbook_id=workbook_id_value, worksheet_id=worksheet_id_value, workstep_id=workstep_id_value, callback=callback_function)

        :param callback function: The callback function
            for asynchronous request. (optional)
        :param str workbook_id: ID of the workbook for which to retrieve the worksteps (required)
        :param str worksheet_id: ID of the worksheet for which to retrieve the worksteps (required)
        :param str workstep_id: ID of the workstep to retrieve (required)
        :return: WorkstepOutputV1
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: WorkstepOutputV1
        """
        kwargs['_return_http_data_only'] = True
        if kwargs.get('callback'):
            return self.get_workstep_with_http_info(**kwargs)
        else:
            (data) = self.get_workstep_with_http_info(**kwargs)
            return data

    def get_workstep_with_http_info(self, **kwargs):
        """
        Get a workstep
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please define a `callback` function
        to be invoked when receiving the response.
        >>> def callback_function(response):
        >>>     pprint(response)
        >>>
        >>> thread = api.get_workstep_with_http_info(workbook_id=workbook_id_value, worksheet_id=worksheet_id_value, workstep_id=workstep_id_value, callback=callback_function)

        :param callback function: The callback function
            for asynchronous request. (optional)
        :param str workbook_id: ID of the workbook for which to retrieve the worksteps (required)
        :param str worksheet_id: ID of the worksheet for which to retrieve the worksteps (required)
        :param str workstep_id: ID of the workstep to retrieve (required)
        :return: WorkstepOutputV1
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: WorkstepOutputV1
        """

        all_params = ['workbook_id', 'worksheet_id', 'workstep_id']
        all_params.append('callback')
        all_params.append('_return_http_data_only')
        all_params.append('_preload_content')
        all_params.append('_request_timeout')

        params = locals()
        for key, val in iteritems(params['kwargs']):
            if key not in all_params:
                raise TypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method get_workstep" % key
                )
            params[key] = val
        del params['kwargs']
        # verify the required parameter 'workbook_id' is set
        if ('workbook_id' not in params) or (params['workbook_id'] is None):
            raise ValueError("Missing the required parameter `workbook_id` when calling `get_workstep`")
        # verify the required parameter 'worksheet_id' is set
        if ('worksheet_id' not in params) or (params['worksheet_id'] is None):
            raise ValueError("Missing the required parameter `worksheet_id` when calling `get_workstep`")
        # verify the required parameter 'workstep_id' is set
        if ('workstep_id' not in params) or (params['workstep_id'] is None):
            raise ValueError("Missing the required parameter `workstep_id` when calling `get_workstep`")


        collection_formats = {}

        path_params = {}
        if 'workbook_id' in params:
            path_params['workbookId'] = params['workbook_id']
        if 'worksheet_id' in params:
            path_params['worksheetId'] = params['worksheet_id']
        if 'workstep_id' in params:
            path_params['workstepId'] = params['workstep_id']

        query_params = []

        header_params = {}

        form_params = []
        local_var_files = {}

        body_params = None
        # HTTP header `Accept`
        header_params['Accept'] = self.api_client.\
            select_header_accept(['application/vnd.seeq.v1+json'])

        # HTTP header `Content-Type`
        header_params['Content-Type'] = self.api_client.\
            select_header_content_type(['application/vnd.seeq.v1+json', ])

        # Authentication setting
        auth_settings = ['api_key']

        return self.api_client.call_api('/workbooks/{workbookId}/worksheets/{worksheetId}/worksteps/{workstepId}', 'GET',
                                        path_params,
                                        query_params,
                                        header_params,
                                        body=body_params,
                                        post_params=form_params,
                                        files=local_var_files,
                                        response_type='WorkstepOutputV1',
                                        auth_settings=auth_settings,
                                        callback=params.get('callback'),
                                        _return_http_data_only=params.get('_return_http_data_only'),
                                        _preload_content=params.get('_preload_content', True),
                                        _request_timeout=params.get('_request_timeout'),
                                        collection_formats=collection_formats)

    def move_worksheet(self, **kwargs):
        """
        Move a Worksheet to a new location in the same Workbook
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please define a `callback` function
        to be invoked when receiving the response.
        >>> def callback_function(response):
        >>>     pprint(response)
        >>>
        >>> thread = api.move_worksheet(workbook_id=workbook_id_value, worksheet_id=worksheet_id_value, callback=callback_function)

        :param callback function: The callback function
            for asynchronous request. (optional)
        :param str workbook_id: ID of the workbook on which the worksheet exists (required)
        :param str worksheet_id: ID of the worksheet on which the workstep exists (required)
        :param str next_worksheet_id: ID of the worksheet that the specified worksheet should precede. If not supplied then the worksheet will be moved to be the last worksheet in the workbook.
        :return: WorksheetOutputV1
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: WorksheetOutputV1
        """
        kwargs['_return_http_data_only'] = True
        if kwargs.get('callback'):
            return self.move_worksheet_with_http_info(**kwargs)
        else:
            (data) = self.move_worksheet_with_http_info(**kwargs)
            return data

    def move_worksheet_with_http_info(self, **kwargs):
        """
        Move a Worksheet to a new location in the same Workbook
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please define a `callback` function
        to be invoked when receiving the response.
        >>> def callback_function(response):
        >>>     pprint(response)
        >>>
        >>> thread = api.move_worksheet_with_http_info(workbook_id=workbook_id_value, worksheet_id=worksheet_id_value, callback=callback_function)

        :param callback function: The callback function
            for asynchronous request. (optional)
        :param str workbook_id: ID of the workbook on which the worksheet exists (required)
        :param str worksheet_id: ID of the worksheet on which the workstep exists (required)
        :param str next_worksheet_id: ID of the worksheet that the specified worksheet should precede. If not supplied then the worksheet will be moved to be the last worksheet in the workbook.
        :return: WorksheetOutputV1
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: WorksheetOutputV1
        """

        all_params = ['workbook_id', 'worksheet_id', 'next_worksheet_id']
        all_params.append('callback')
        all_params.append('_return_http_data_only')
        all_params.append('_preload_content')
        all_params.append('_request_timeout')

        params = locals()
        for key, val in iteritems(params['kwargs']):
            if key not in all_params:
                raise TypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method move_worksheet" % key
                )
            params[key] = val
        del params['kwargs']
        # verify the required parameter 'workbook_id' is set
        if ('workbook_id' not in params) or (params['workbook_id'] is None):
            raise ValueError("Missing the required parameter `workbook_id` when calling `move_worksheet`")
        # verify the required parameter 'worksheet_id' is set
        if ('worksheet_id' not in params) or (params['worksheet_id'] is None):
            raise ValueError("Missing the required parameter `worksheet_id` when calling `move_worksheet`")


        collection_formats = {}

        path_params = {}
        if 'workbook_id' in params:
            path_params['workbookId'] = params['workbook_id']
        if 'worksheet_id' in params:
            path_params['worksheetId'] = params['worksheet_id']

        query_params = []
        if 'next_worksheet_id' in params:
            query_params.append(('nextWorksheetId', params['next_worksheet_id']))

        header_params = {}

        form_params = []
        local_var_files = {}

        body_params = None
        # HTTP header `Accept`
        header_params['Accept'] = self.api_client.\
            select_header_accept(['application/vnd.seeq.v1+json'])

        # HTTP header `Content-Type`
        header_params['Content-Type'] = self.api_client.\
            select_header_content_type(['application/vnd.seeq.v1+json', ])

        # Authentication setting
        auth_settings = ['api_key']

        return self.api_client.call_api('/workbooks/{workbookId}/worksheets/{worksheetId}/next', 'POST',
                                        path_params,
                                        query_params,
                                        header_params,
                                        body=body_params,
                                        post_params=form_params,
                                        files=local_var_files,
                                        response_type='WorksheetOutputV1',
                                        auth_settings=auth_settings,
                                        callback=params.get('callback'),
                                        _return_http_data_only=params.get('_return_http_data_only'),
                                        _preload_content=params.get('_preload_content', True),
                                        _request_timeout=params.get('_request_timeout'),
                                        collection_formats=collection_formats)

    def set_current_workstep(self, **kwargs):
        """
        Set the current workstep
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please define a `callback` function
        to be invoked when receiving the response.
        >>> def callback_function(response):
        >>>     pprint(response)
        >>>
        >>> thread = api.set_current_workstep(workbook_id=workbook_id_value, worksheet_id=worksheet_id_value, workstep_id=workstep_id_value, callback=callback_function)

        :param callback function: The callback function
            for asynchronous request. (optional)
        :param str workbook_id: ID of the workbook on which the worksheet exists (required)
        :param str worksheet_id: ID of the worksheet on which the workstep exists (required)
        :param str workstep_id: ID of the workstep to set as the current one (required)
        :return: WorkstepOutputV1
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: WorkstepOutputV1
        """
        kwargs['_return_http_data_only'] = True
        if kwargs.get('callback'):
            return self.set_current_workstep_with_http_info(**kwargs)
        else:
            (data) = self.set_current_workstep_with_http_info(**kwargs)
            return data

    def set_current_workstep_with_http_info(self, **kwargs):
        """
        Set the current workstep
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please define a `callback` function
        to be invoked when receiving the response.
        >>> def callback_function(response):
        >>>     pprint(response)
        >>>
        >>> thread = api.set_current_workstep_with_http_info(workbook_id=workbook_id_value, worksheet_id=worksheet_id_value, workstep_id=workstep_id_value, callback=callback_function)

        :param callback function: The callback function
            for asynchronous request. (optional)
        :param str workbook_id: ID of the workbook on which the worksheet exists (required)
        :param str worksheet_id: ID of the worksheet on which the workstep exists (required)
        :param str workstep_id: ID of the workstep to set as the current one (required)
        :return: WorkstepOutputV1
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: WorkstepOutputV1
        """

        all_params = ['workbook_id', 'worksheet_id', 'workstep_id']
        all_params.append('callback')
        all_params.append('_return_http_data_only')
        all_params.append('_preload_content')
        all_params.append('_request_timeout')

        params = locals()
        for key, val in iteritems(params['kwargs']):
            if key not in all_params:
                raise TypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method set_current_workstep" % key
                )
            params[key] = val
        del params['kwargs']
        # verify the required parameter 'workbook_id' is set
        if ('workbook_id' not in params) or (params['workbook_id'] is None):
            raise ValueError("Missing the required parameter `workbook_id` when calling `set_current_workstep`")
        # verify the required parameter 'worksheet_id' is set
        if ('worksheet_id' not in params) or (params['worksheet_id'] is None):
            raise ValueError("Missing the required parameter `worksheet_id` when calling `set_current_workstep`")
        # verify the required parameter 'workstep_id' is set
        if ('workstep_id' not in params) or (params['workstep_id'] is None):
            raise ValueError("Missing the required parameter `workstep_id` when calling `set_current_workstep`")


        collection_formats = {}

        path_params = {}
        if 'workbook_id' in params:
            path_params['workbookId'] = params['workbook_id']
        if 'worksheet_id' in params:
            path_params['worksheetId'] = params['worksheet_id']
        if 'workstep_id' in params:
            path_params['workstepId'] = params['workstep_id']

        query_params = []

        header_params = {}

        form_params = []
        local_var_files = {}

        body_params = None
        # HTTP header `Accept`
        header_params['Accept'] = self.api_client.\
            select_header_accept(['application/vnd.seeq.v1+json'])

        # HTTP header `Content-Type`
        header_params['Content-Type'] = self.api_client.\
            select_header_content_type(['application/vnd.seeq.v1+json', ])

        # Authentication setting
        auth_settings = ['api_key']

        return self.api_client.call_api('/workbooks/{workbookId}/worksheets/{worksheetId}/worksteps/{workstepId}/current', 'POST',
                                        path_params,
                                        query_params,
                                        header_params,
                                        body=body_params,
                                        post_params=form_params,
                                        files=local_var_files,
                                        response_type='WorkstepOutputV1',
                                        auth_settings=auth_settings,
                                        callback=params.get('callback'),
                                        _return_http_data_only=params.get('_return_http_data_only'),
                                        _preload_content=params.get('_preload_content', True),
                                        _request_timeout=params.get('_request_timeout'),
                                        collection_formats=collection_formats)

    def set_renderer(self, **kwargs):
        """
        Set the topic's renderer.
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please define a `callback` function
        to be invoked when receiving the response.
        >>> def callback_function(response):
        >>>     pprint(response)
        >>>
        >>> thread = api.set_renderer(id=id_value, identity_id=identity_id_value, callback=callback_function)

        :param callback function: The callback function
            for asynchronous request. (optional)
        :param str id: ID of the topic to change the renderer of (required)
        :param str identity_id: The user or group to make the topic's renderer (required)
        :return: StatusMessageBase
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: StatusMessageBase
        """
        kwargs['_return_http_data_only'] = True
        if kwargs.get('callback'):
            return self.set_renderer_with_http_info(**kwargs)
        else:
            (data) = self.set_renderer_with_http_info(**kwargs)
            return data

    def set_renderer_with_http_info(self, **kwargs):
        """
        Set the topic's renderer.
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please define a `callback` function
        to be invoked when receiving the response.
        >>> def callback_function(response):
        >>>     pprint(response)
        >>>
        >>> thread = api.set_renderer_with_http_info(id=id_value, identity_id=identity_id_value, callback=callback_function)

        :param callback function: The callback function
            for asynchronous request. (optional)
        :param str id: ID of the topic to change the renderer of (required)
        :param str identity_id: The user or group to make the topic's renderer (required)
        :return: StatusMessageBase
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: StatusMessageBase
        """

        all_params = ['id', 'identity_id']
        all_params.append('callback')
        all_params.append('_return_http_data_only')
        all_params.append('_preload_content')
        all_params.append('_request_timeout')

        params = locals()
        for key, val in iteritems(params['kwargs']):
            if key not in all_params:
                raise TypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method set_renderer" % key
                )
            params[key] = val
        del params['kwargs']
        # verify the required parameter 'id' is set
        if ('id' not in params) or (params['id'] is None):
            raise ValueError("Missing the required parameter `id` when calling `set_renderer`")
        # verify the required parameter 'identity_id' is set
        if ('identity_id' not in params) or (params['identity_id'] is None):
            raise ValueError("Missing the required parameter `identity_id` when calling `set_renderer`")


        collection_formats = {}

        path_params = {}
        if 'id' in params:
            path_params['id'] = params['id']

        query_params = []
        if 'identity_id' in params:
            query_params.append(('identityId', params['identity_id']))

        header_params = {}

        form_params = []
        local_var_files = {}

        body_params = None
        # HTTP header `Accept`
        header_params['Accept'] = self.api_client.\
            select_header_accept(['application/vnd.seeq.v1+json'])

        # HTTP header `Content-Type`
        header_params['Content-Type'] = self.api_client.\
            select_header_content_type(['application/vnd.seeq.v1+json', ])

        # Authentication setting
        auth_settings = ['api_key']

        return self.api_client.call_api('/workbooks/{id}/renderer', 'POST',
                                        path_params,
                                        query_params,
                                        header_params,
                                        body=body_params,
                                        post_params=form_params,
                                        files=local_var_files,
                                        response_type='StatusMessageBase',
                                        auth_settings=auth_settings,
                                        callback=params.get('callback'),
                                        _return_http_data_only=params.get('_return_http_data_only'),
                                        _preload_content=params.get('_preload_content', True),
                                        _request_timeout=params.get('_request_timeout'),
                                        collection_formats=collection_formats)
